@@ -4,7 +4,7 @@
 #     All rights reserved.
 #     This is free software with ABSOLUTELY NO WARRANTY.
 #
-# You can redistribute it and/or modify it under the terms of 
+# You can redistribute it and/or modify it under the terms of
 # the GNU General Public License version 2.
 
 require 'senna_api'
@@ -77,7 +77,7 @@ module Senna
   LOG_TITLE                     =      0x0002
   LOG_MESSAGE                   =      0x0004
   LOG_LOCATION                  =      0x0008
-  
+
   class Index
     attr_reader :handle
     attr_reader :key_size, :flags, :initial_n_segments, :encoding
@@ -108,7 +108,8 @@ module Senna
       id = [id].pack('L') if Fixnum === id
       oldvalue ||= ''
       newvalue ||= ''
-      Senna::sen_index_upd(@handle, id, oldvalue, newvalue)
+      Senna::sen_index_upd(@handle, id, oldvalue, oldvalue.length,
+                           newvalue, newvalue.length)
     end
 
     def update(id, section, oldvalue, newvalue)
@@ -122,19 +123,19 @@ module Senna
     end
 
     def sel(query)
-      r = Senna::sen_index_sel(@handle, query)
+      r = Senna::sen_index_sel(@handle, query, query.length)
       return Records.new(r) if r
     end
 
     def select(query, records=nil, op=SEL_OR, optarg=nil)
       records ||= Records.new
-      Senna::sen_index_select(@handle, query, records.handle, op, optarg)
+      Senna::sen_index_select(@handle, query, query.length, records.handle, op, optarg)
       return records
     end
 
     def query(query, default_op=SEL_OR, max_exprs=32, encoding=ENC_DEFAULT)
-      return unless String === query 
-      q = Senna::sen_query_open(query, default_op, max_exprs, encoding)
+      return unless String === query
+      q = Senna::sen_query_open(query, query.length, default_op, max_exprs, encoding)
       r = Senna::sen_records_open(REC_DOCUMENT, REC_NONE, 0)
       Senna::sen_query_exec(@handle, q, r, SEL_OR)
       Senna::sen_query_close(q)
@@ -161,7 +162,7 @@ module Senna
     end
 
     def path
-      size, = Senna::sen_index_path(@handle, 0)
+      size, = Senna::sen_index_path(@handle, 1)
       _,res = Senna::sen_index_path(@handle, size)
       return res
     end
@@ -268,12 +269,12 @@ module Senna
         Senna::sen_inv_entry_info(index.inv, term_id)
     end
   end
-  
+
   class Sym
     def Sym.callback(handle)
       proc { Senna::sen_sym_close(handle) }
     end
-    
+
     def Sym::create(path, key_size=0, flags=0, encoding=0)
       handle = Senna::sen_sym_create(path, key_size, flags, encoding)
       return new(handle, path)
@@ -285,7 +286,7 @@ module Senna
     end
 
     def index_open_with_keys(path)
-      handle = Senna::sen_index_open_with_keys(path, @handle) 
+      handle = Senna::sen_index_open_with_keys(path, @handle)
       return Senna::Index.new(handle)
     end
 
@@ -293,7 +294,7 @@ module Senna
       handle = Senna::sen_sym_open(path)
       return new(handle, path)
     end
-    
+
     def initialize(handle, path=nil)
       @handle = handle
       @path = path
@@ -322,7 +323,7 @@ module Senna
     def get(key)
       Senna::sen_sym_get(@handle, key)
     end
-    
+
     def at(key)
       Senna::sen_sym_at(@handle, key)
     end
@@ -384,7 +385,7 @@ module Senna
     def Set.callback(handle)
       proc { Senna::sen_set_close(handle) }
     end
-    
+
     def Set::open(keysize, value_size, index_size)
       handle = Senna::sen_set_open(keysize, value_size, index_size);
       return new(handle)
@@ -426,7 +427,7 @@ module Senna
     def cursor_open
       Senna::sen_set_cursor_open(@handle)
     end
-    
+
     def cursor_next(cursor)
       eh, _, _ = Senna::sen_set_cursor_next(cursor)
       return eh
@@ -435,7 +436,7 @@ module Senna
     def cursor_close(cursor)
       Senna::sen_set_cursor_close(cursor)
     end
-    
+
     def element_info(eh, key=0, value=0)
       Senna::sen_set_element_info(@handle, eh, key, value)
     end
@@ -443,7 +444,7 @@ module Senna
     def +(obj)
       union(obj)
     end
-    
+
     def union(obj)
       Senna::sen_set_union(@handle, obj)
     end
@@ -602,7 +603,7 @@ module Senna
     def initialize(phandle, handle, bufsize=0)
       @phandle = phandle
       @handle = handle
-      rc, @key, @keysize, @section, @pos, @score, @n_subrecs = 
+      rc, @key, @keysize, @section, @pos, @score, @n_subrecs =
         Senna::sen_record_info(@phandle, @handle, bufsize)
     end
 
@@ -636,7 +637,7 @@ module Senna
     end
 
     def add(str, weight=1)
-      Senna::sen_values_add(@handle, str, weight)
+      Senna::sen_values_add(@handle, str, str.length, weight)
     end
 
     def initialize(handle)
@@ -659,7 +660,7 @@ module Senna
     end
 
     def Query::open(str, default_op=SEL_OR, max_exprs=32, encoding=ENC_DEFAULT)
-      handle = Senna::sen_query_open(str, default_op, max_exprs, encoding)
+      handle = Senna::sen_query_open(str, str.length, default_op, max_exprs, encoding)
       return new(handle)
     end
 
@@ -706,8 +707,17 @@ module Senna
       else
         condkeys = []
       end
-      handle = Senna::sen_snip_open(enc, flags, width, max_results, opentag, closetag, map)
-      condkeys.each {|c| Senna::sen_snip_add_cond(handle, c, condvals[c][0], condvals[c][1]) }
+      handle = Senna::sen_snip_open(enc, flags, width, max_results,
+                                    opentag, opentag.length,
+                                    closetag, closetag.length,
+                                    map)
+      condkeys.each {|c|
+        opentag_len = condvals[c][0].nil? ? 0 : condvals[c][0].length
+        closetag_len = condvals[c][1].nil? ? 0 : condvals[c][1].length
+        Senna::sen_snip_add_cond(handle, c, c.length,
+                                 condvals[c][0], opentag_len,
+                                 condvals[c][1], closetag_len)
+      }
       return Snip.new(handle)
     end
 
@@ -717,15 +727,19 @@ module Senna
     end
 
     def add(cond, opentag=nil, closetag=nil)
-      Senna::sen_snip_add_cond(@handle, cond, opentag, closetag)
+      opentag_len = opentag.nil? ? 0 : opentag.length
+      closetag_len = closetag.nil? ? 0 : closetag.length
+      Senna::sen_snip_add_cond(@handle, cond, cond.length,
+                               opentag, opentag_len,
+                               closetag, closetag_len)
     end
 
     def exec(str)
-      rc, n, maxsize = Senna::sen_snip_exec(@handle, str)
+      rc, n, maxsize = Senna::sen_snip_exec(@handle, str, str.length)
       res = []
       for i in 0..n-1
         rc, r = Senna::sen_snip_get_result(@handle, i, maxsize)
-	res << r if rc == 0
+        res << r if rc == 0
       end
       return res
     end
@@ -735,9 +749,9 @@ module Senna
       @handle = nil
       ObjectSpace.undefine_finalizer(self)
       return rc
-    end    
+    end
   end
-  
+
   def Senna::get_select_optarg(mode, similarity_threshold_max_interval=0, weight_vector=nil, &func)
     i = Senna::SelectOptarg.new
     i.mode = mode
