@@ -58,7 +58,7 @@ cursor_heap_push(cursor_heap *h, sen_inv *inv, sen_id tid, uint32_t offset2)
   if (h->n_entries >= h->n_bins) {
     int max = h->n_bins * 2;
     sen_inv_cursor **bins = SEN_REALLOC(h->bins, sizeof(sen_inv_cursor *) * max);
-    sen_log("expanded cursor_heap to %d,%p", max, bins);
+    SEN_LOG(sen_log_debug, "expanded cursor_heap to %d,%p", max, bins);
     if (!bins) { return sen_memory_exhausted; }
     h->n_bins = max;
     h->bins = bins;
@@ -66,15 +66,15 @@ cursor_heap_push(cursor_heap *h, sen_inv *inv, sen_id tid, uint32_t offset2)
 #ifdef USE_AIO
   if (sen_aio_enabled) {
     if (!(c = sen_inv_cursor_openv1(inv, tid))) {
-      sen_log("cursor open failed");
+      SEN_LOG(sen_log_error, "cursor open failed");
       return sen_other_error;
     }
     h->bins[h->n_entries++] = c;
   } else
 #endif /* USE_AIO */
   {
-    if (!(c = sen_inv_cursor_open(inv, tid))) {
-      sen_log("cursor open failed");
+    if (!(c = sen_inv_cursor_open(inv, tid, 1))) {
+      SEN_LOG(sen_log_error, "cursor open failed");
       return sen_other_error;
     }
     if (sen_inv_cursor_next(c)) {
@@ -82,7 +82,7 @@ cursor_heap_push(cursor_heap *h, sen_inv *inv, sen_id tid, uint32_t offset2)
       return sen_other_error;
     }
     if (sen_inv_cursor_next_pos(c)) {
-      sen_log("invalid inv_cursor b");
+      SEN_LOG(sen_log_error, "invalid inv_cursor b");
       sen_inv_cursor_close(c);
       return sen_other_error;
     }
@@ -104,11 +104,13 @@ static inline sen_rc
 cursor_heap_push2(cursor_heap *h)
 {
 #ifdef USE_AIO
+  sen_rc rc;
   if (sen_aio_enabled) {
     int i, j, n, n2;
     sen_inv_cursor *c, *c2;
     if (h && h->n_entries) {
-      sen_inv_cursor_openv2(h->bins, h->n_entries);
+      rc = sen_inv_cursor_openv2(h->bins, h->n_entries);
+      SEN_ASSERT(rc);
       for (i = 0, j = 0; i < h->n_entries; i++) {
         c = h->bins[i];
         if (sen_inv_cursor_next(c)) {
@@ -116,7 +118,7 @@ cursor_heap_push2(cursor_heap *h)
           continue;
         }
         if (sen_inv_cursor_next_pos(c)) {
-          sen_log("invalid inv_cursor b");
+          SEN_LOG(sen_log_error, "invalid inv_cursor b");
           sen_inv_cursor_close(c);
           continue;
         }
@@ -184,7 +186,7 @@ cursor_heap_pop(cursor_heap *h)
       sen_inv_cursor_close(c);
       h->bins[0] = h->bins[--h->n_entries];
     } else if (sen_inv_cursor_next_pos(c)) {
-      sen_log("invalid inv_cursor c");
+      SEN_LOG(sen_log_error, "invalid inv_cursor c");
       sen_inv_cursor_close(c);
       h->bins[0] = h->bins[--h->n_entries];
     }
@@ -202,7 +204,7 @@ cursor_heap_pop_pos(cursor_heap *h)
         sen_inv_cursor_close(c);
         h->bins[0] = h->bins[--h->n_entries];
       } else if (sen_inv_cursor_next_pos(c)) {
-        sen_log("invalid inv_cursor d");
+        SEN_LOG(sen_log_error, "invalid inv_cursor d");
         sen_inv_cursor_close(c);
         h->bins[0] = h->bins[--h->n_entries];
       }
@@ -250,7 +252,15 @@ token_info_expand_both(sen_index *i, const char *key, token_info *ti)
   if (!(h = sen_sym_prefix_search(i->lexicon, key))) { return; }
   // sen_log("key=%s h->n=%d", key, h->n_entries);
   ti->cursors = cursor_heap_open(h->n_entries + 256);
+  SEN_ASSERT(ti->cursors);
+/*
+  if (!ti->cursors) { sen_set_close(h); return; }
+*/
   c = sen_set_cursor_open(h);
+  SEN_ASSERT(c);
+/*
+  if (!c) { sen_set_close(h); return; }
+*/
   while (sen_set_cursor_next(c, (void **) &tp, NULL)) {
     const char *key2 = _sen_sym_key(i->lexicon, *tp);
     // sen_log("key2=%s", key2);
@@ -265,6 +275,10 @@ token_info_expand_both(sen_index *i, const char *key, token_info *ti)
       // sen_log("g=%d", g ? g->n_entries : 0);
       if (g) {
         sen_set_cursor *d = sen_set_cursor_open(g);
+        SEN_ASSERT(d);
+/*
+        if (!d) { sen_set_close(g); sen_set_cursor_close(c); sen_set_close(h); return; }
+*/
         while (sen_set_cursor_next(d, (void **) &tq, (void **) &offset2)) {
           if ((s = sen_inv_estimate_size(i->inv, *tq))) {
             // sen_log("est=%d key=%s", s, _sen_sym_key(i->lexicon, *tq));
@@ -301,6 +315,7 @@ token_info_open(sen_index *i, const char *key, uint32_t offset, expand_mode mode
   sen_id tid;
   sen_set_cursor *c;
   sen_id *tp;
+  sen_rc rc;
   if (!(ti = SEN_MALLOC(sizeof(token_info)))) { return NULL; }
   ti->cursors = NULL;
   ti->size = 0;
@@ -314,6 +329,10 @@ token_info_open(sen_index *i, const char *key, uint32_t offset, expand_mode mode
     if (!(tid = sen_sym_at(i->lexicon, key))) { break; }
     if (!(s = sen_inv_estimate_size(i->inv, tid))) { break; }
     ti->cursors = cursor_heap_open(1);
+    SEN_ASSERT(ti->cursors);
+/*
+    if (!ti->cursors) { SEN_FREE(ti); return NULL; }
+*/
     cursor_heap_push(ti->cursors, i->inv, tid, 0);
     ti->ntoken++;
     ti->size = s;
@@ -322,7 +341,15 @@ token_info_open(sen_index *i, const char *key, uint32_t offset, expand_mode mode
     if (!(h = sen_sym_prefix_search(i->lexicon, key))) { break; }
     // sen_log("key=%s h->n=%d", key, h->n_entries);
     ti->cursors = cursor_heap_open(h->n_entries);
+    SEN_ASSERT(ti->cursors);
+/*
+    if (!ti->cursors) { sen_set_close(h); SEN_FREE(ti); return NULL; }
+*/
     c = sen_set_cursor_open(h);
+    SEN_ASSERT(c);
+/*
+    if (!c) { sen_set_close(h); SEN_FREE(ti->cursors); SEN_FREE(ti); return NULL; }
+*/
     while (sen_set_cursor_next(c, (void **) &tp, NULL)) {
       if ((s = sen_inv_estimate_size(i->inv, *tp))) {
         // sen_log("%8d %s", s, _sen_sym_key(i->lexicon, *tp));
@@ -342,7 +369,15 @@ token_info_open(sen_index *i, const char *key, uint32_t offset, expand_mode mode
       if (!(h = sen_sym_suffix_search(i->lexicon, key))) { break; }
       // sen_log("key=%s h->n=%d", key, h->n_entries);
       ti->cursors = cursor_heap_open(h->n_entries);
+      SEN_ASSERT(ti->cursors);
+/*
+      if (!ti->cursors) { sen_set_close(h); SEN_FREE(ti); return NULL; }
+*/ 
       c = sen_set_cursor_open(h);
+      SEN_ASSERT(c);
+/*
+      if (!c) { cursor_heap_close(ti->cursors); sen_set_close(h); SEN_FREE(ti); return NULL; }
+*/
       while (sen_set_cursor_next(c, (void **) &tp, (void **) &offset2)) {
         if ((s = sen_inv_estimate_size(i->inv, *tp))) {
           cursor_heap_push(ti->cursors, i->inv, *tp, /* *offset2 */ 0);
@@ -355,7 +390,11 @@ token_info_open(sen_index *i, const char *key, uint32_t offset, expand_mode mode
       break;
     }
   }
-  cursor_heap_push2(ti->cursors);
+  rc = cursor_heap_push2(ti->cursors);
+  SEN_ASSERT(!rc);
+/*
+  if (rc) { SEN_FREE(ti); return NULL; }
+*/
   {
     sen_inv_cursor *ic;
     if (ti->cursors && (ic = cursor_heap_min(ti->cursors))) {
@@ -493,11 +532,11 @@ expand_stack(void)
 #ifndef WIN32
   struct rlimit rlim;
   if (!getrlimit(RLIMIT_STACK, &rlim)) {
-    sen_log("RLIMIT_STACK is %d (%d)", rlim.rlim_cur, getpid());
+    SEN_LOG(sen_log_notice, "RLIMIT_STACK is %d (%d)", rlim.rlim_cur, getpid());
     if (rlim.rlim_cur != RLIM_INFINITY && rlim.rlim_cur < SEN_STACK_SIZE) {
       rlim.rlim_cur = SEN_STACK_SIZE;
       if (!setrlimit(RLIMIT_STACK, &rlim)) {
-        sen_log("expanded RLIMIT_STACK to %d", rlim.rlim_cur);
+        SEN_LOG(sen_log_notice, "expanded RLIMIT_STACK to %d", rlim.rlim_cur);
       }
     }
   }
@@ -518,14 +557,14 @@ read_conf(void)
         if (p - buffer > BUFSIZE) { break; }
       }
       sen_initial_n_segments = atoi(p);
-      sen_log("INITIAL_N_SEGMENTS=%d", sen_initial_n_segments);
+      SEN_LOG(sen_log_notice, "INITIAL_N_SEGMENTS=%d", sen_initial_n_segments);
     }
     if (!strncmp(buffer, partial_match_threshold, strlen(partial_match_threshold))) {
       for (p = buffer; *p && (*p < '0' || *p > '9') && *p != '-'; p++) {
         if (p - buffer > BUFSIZE) { break; }
       }
       sen_query_escalation_threshold = atoi(p);
-      sen_log("PARTIAL_MATCH_THRESHOLD=%d", sen_query_escalation_threshold);
+      SEN_LOG(sen_log_notice, "PARTIAL_MATCH_THRESHOLD=%d", sen_query_escalation_threshold);
     }
     if (!strncmp(buffer, default_encoding, strlen(default_encoding))) {
       for (p = buffer + strlen(default_encoding); *p; p++) {
@@ -579,7 +618,7 @@ sen_init(void)
     sen_aio_enabled = 0;
   }
   if (sen_aio_enabled) {
-    sen_log("AIO and DIO enabled");
+    SEN_LOG(sen_log_notice, "AIO and DIO enabled");
     sen_cache_open();
   }
 #endif /* USE_AIO */
@@ -624,14 +663,15 @@ sen_index_create(const char *path, int key_size,
   sen_index *i;
   char buffer[PATH_MAX];
   read_conf();
+  if (!path) { SEN_LOG(sen_log_warning, "sen_index_create: invalid argument"); return NULL; }
   if (initial_n_segments == 0) { initial_n_segments = sen_initial_n_segments; }
   if (encoding == sen_enc_default) { encoding = sen_default_encoding; }
   if (strlen(path) > PATH_MAX - 4) {
-    sen_log("too long index path (%s)", path);
+    SEN_LOG(sen_log_warning, "sen_index_create: too long index path (%s)", path);
     return NULL;
   }
   if (!(i = SEN_MALLOC(sizeof(sen_index)))) { return NULL; }
-  sen_log("creating '%s' encoding=%s initial_n_segments=%d",
+  SEN_LOG(sen_log_notice, "creating '%s' encoding=%s initial_n_segments=%d",
           path, sen_enc_string[encoding], initial_n_segments);
   strcpy(buffer, path);
   strcat(buffer, ".SEN");
@@ -643,18 +683,18 @@ sen_index_create(const char *path, int key_size,
       strcpy(buffer, path);
       strcat(buffer, ".SEN.i");
       if ((i->inv = sen_inv_create(buffer, i->lexicon, 0, initial_n_segments))) {
-	if ((flags & SEN_INDEX_WITH_STORE)) {
-	  strcpy(buffer, path);
-	  strcat(buffer, ".SEN.v");
-	  i->store= sen_store_create(buffer);
-	} else {
-	  i->store = NULL;
-	}
-	if (!(flags & SEN_INDEX_WITH_STORE) || i->store) {
-	  sen_log("index created (%s) flags=%x", path, i->lexicon->flags);
-	  return i;
-	}
-	sen_inv_close(i->inv);
+        if ((flags & SEN_INDEX_WITH_STORE)) {
+          strcpy(buffer, path);
+          strcat(buffer, ".SEN.v");
+          i->store= sen_store_create(buffer);
+        } else {
+          i->store = NULL;
+        }
+        if (!(flags & SEN_INDEX_WITH_STORE) || i->store) {
+          SEN_LOG(sen_log_notice, "index created (%s) flags=%x", path, i->lexicon->flags);
+          return i;
+        }
+        sen_inv_close(i->inv);
       }
       sen_sym_close(i->lexicon);
     }
@@ -669,8 +709,9 @@ sen_index_open(const char *path)
 {
   sen_index *i;
   char buffer[PATH_MAX];
+  if (!path) { SEN_LOG(sen_log_warning, "sen_index_open: invalid argument"); return NULL; }
   if (strlen(path) > PATH_MAX - 4) {
-    sen_log("too long index path (%s)", path);
+    SEN_LOG(sen_log_warning, "sen_index_open: too long index path (%s)", path);
     return NULL;
   }
   read_conf();
@@ -685,18 +726,18 @@ sen_index_open(const char *path)
       strcpy(buffer, path);
       strcat(buffer, ".SEN.i");
       if ((i->inv = sen_inv_open(buffer, i->lexicon))) {
-	if ((i->lexicon->flags & SEN_INDEX_WITH_STORE)) {
-	  strcpy(buffer, path);
-	  strcat(buffer, ".SEN.v");
-	  i->store = sen_store_open(buffer);
-	} else {
-	  i->store = NULL;
-	}
-	if(!(i->lexicon->flags & SEN_INDEX_WITH_STORE) || i->store) {
-	  sen_log("index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
-	  return i;
-	}
-	sen_inv_close(i->inv);
+        if ((i->lexicon->flags & SEN_INDEX_WITH_STORE)) {
+          strcpy(buffer, path);
+          strcat(buffer, ".SEN.v");
+          i->store = sen_store_open(buffer);
+        } else {
+          i->store = NULL;
+        }
+        if(!(i->lexicon->flags & SEN_INDEX_WITH_STORE) || i->store) {
+          SEN_LOG(sen_log_notice, "index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
+          return i;
+        }
+        sen_inv_close(i->inv);
       }
       sen_sym_close(i->lexicon);
     }
@@ -713,14 +754,18 @@ sen_index_create_with_keys(const char *path, sen_sym *keys,
   sen_index *i;
   char buffer[PATH_MAX];
   read_conf();
+  if (!path || !keys) {
+    SEN_LOG(sen_log_warning, "sen_index_create_with_keys: invalid argument");
+    return NULL;
+  }
   if (initial_n_segments == 0) { initial_n_segments = sen_initial_n_segments; }
   if (encoding == sen_enc_default) { encoding = sen_default_encoding; }
   if (strlen(path) > PATH_MAX - 4) {
-    sen_log("too long index path (%s)", path);
+    SEN_LOG(sen_log_warning, "too long index path (%s)", path);
     return NULL;
   }
   if (!(i = SEN_MALLOC(sizeof(sen_index)))) { return NULL; }
-  sen_log("creating '%s' encoding=%s initial_n_segments=%d",
+  SEN_LOG(sen_log_notice, "creating '%s' encoding=%s initial_n_segments=%d",
           path, sen_enc_string[encoding], initial_n_segments);
   i->keys = keys;
   i->foreign_keyp = 1;
@@ -731,15 +776,15 @@ sen_index_create_with_keys(const char *path, sen_sym *keys,
     strcat(buffer, ".SEN.i");
     if ((i->inv = sen_inv_create(buffer, i->lexicon, 0, initial_n_segments))) {
       if ((flags & SEN_INDEX_WITH_STORE)) {
-	strcpy(buffer, path);
-	strcat(buffer, ".SEN.v");
-	i->store= sen_store_create(buffer);
+        strcpy(buffer, path);
+        strcat(buffer, ".SEN.v");
+        i->store= sen_store_create(buffer);
       } else {
-	i->store = NULL;
+        i->store = NULL;
       }
       if (!(flags & SEN_INDEX_WITH_STORE) || i->store) {
-	sen_log("index created (%s) flags=%x", path, i->lexicon->flags);
-	return i;
+        SEN_LOG(sen_log_notice, "index created (%s) flags=%x", path, i->lexicon->flags);
+        return i;
       }
       sen_inv_close(i->inv);
     }
@@ -754,8 +799,12 @@ sen_index_open_with_keys(const char *path, sen_sym *keys)
 {
   sen_index *i;
   char buffer[PATH_MAX];
+  if (!path || !keys) {
+    SEN_LOG(sen_log_warning, "sen_index_open_with_keys: invalid argument");
+    return NULL;
+  }
   if (strlen(path) > PATH_MAX - 4) {
-    sen_log("too long index path (%s)", path);
+    SEN_LOG(sen_log_warning, "too long index path (%s)", path);
     return NULL;
   }
   read_conf();
@@ -769,15 +818,15 @@ sen_index_open_with_keys(const char *path, sen_sym *keys)
     strcat(buffer, ".SEN.i");
     if ((i->inv = sen_inv_open(buffer, i->lexicon))) {
       if ((i->lexicon->flags & SEN_INDEX_WITH_STORE)) {
-	strcpy(buffer, path);
-	strcat(buffer, ".SEN.v");
-	i->store = sen_store_open(buffer);
+        strcpy(buffer, path);
+        strcat(buffer, ".SEN.v");
+        i->store = sen_store_open(buffer);
       } else {
-	i->store = NULL;
+        i->store = NULL;
       }
       if(!(i->lexicon->flags & SEN_INDEX_WITH_STORE) || i->store) {
-	sen_log("index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
-	return i;
+        SEN_LOG(sen_log_notice, "index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
+        return i;
       }
       sen_inv_close(i->inv);
     }
@@ -857,16 +906,10 @@ sen_index_info(sen_index *i, int *key_size, int *flags,
   if (initial_n_segments) { *initial_n_segments = sen_inv_initial_n_segments(i->inv); }
   if (encoding) { *encoding = i->lexicon->encoding; }
   if (nrecords_keys || file_size_keys) {
-    rc = sen_sym_info(i->keys, NULL, NULL, NULL, nrecords_keys, file_size_keys);
-    if (rc != sen_success) {
-      return rc;
-    }
+    if ((rc = sen_sym_info(i->keys, NULL, NULL, NULL, nrecords_keys, file_size_keys))) { return rc; }
   }
   if (nrecords_lexicon || file_size_lexicon) {
-    rc = sen_sym_info(i->lexicon, NULL, NULL, NULL, nrecords_lexicon, file_size_lexicon);
-    if (rc != sen_success) {
-      return rc;
-    }
+    if ((rc = sen_sym_info(i->lexicon, NULL, NULL, NULL, nrecords_lexicon, file_size_lexicon))) { return rc; }
   }
   if (inv_seg_size || inv_chunk_size) {
     off_t seg_size, chunk_size;
@@ -926,9 +969,16 @@ sen_index_path(sen_index *i, char *pathbuf, int bufsize)
 {
   const char *invpath;
   int pathsize;
-  invpath = sen_io_path(i->inv->seg);
+  SEN_ASSERT(i);
+/*
+  if (!i) {
+    SEN_LOG(sen_log_warning, "sen_index_path: invalid argument");
+    return sen_invalid_argument;
+  }
+*/
+  invpath = sen_io_path(i->lexicon->io);
   pathsize = strlen(invpath) - 5;
-  if (bufsize >= pathsize) {
+  if (bufsize >= pathsize && pathbuf) {
     memcpy(pathbuf, invpath, pathsize - 1);
     pathbuf[pathsize - 1] = '\0';
   }
@@ -956,18 +1006,38 @@ index_add(sen_index *i, const void *key, const char *value)
   }
   if (i->store) { sbuf = sen_store_buf_open(len); }
   h = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
+  if (!h) {
+    SEN_LOG(sen_log_alert, "sen_set_open on index_add failed !");
+    sen_lex_close(lex);
+    return sen_memory_exhausted; 
+  }
   while (!lex->status) {
     (tid = sen_lex_next(lex));
     if (tid) {
       sen_set_get(h, &tid, (void **) &u);
-      if (!*u) { *u = sen_inv_updspec_open(rid, 1); }
-      sen_inv_updspec_add(*u, lex->pos, 0);
+      if (!*u) {
+        if (!(*u = sen_inv_updspec_open(rid, 1))) {
+          SEN_LOG(sen_log_error, "sen_inv_updspec_open on index_add failed!");
+          goto exit;
+        }
+      }
+      if (sen_inv_updspec_add(*u, lex->pos, 0)) {
+        SEN_LOG(sen_log_error, "sen_inv_updspec_add on index_add failed!");
+        goto exit;
+      }
       if (sbuf) { sen_store_buf_add(sbuf, tid); }
     }
   }
   sen_lex_close(lex);
   if (sbuf) { sen_store_update(i->store, rid, sbuf, h); }
   c = sen_set_cursor_open(h);
+  SEN_ASSERT(c);
+/*
+  if (!c) {
+    // fixme: free inv_updspec arrays *u
+    sen_set_close(h); return sen_memory_exhausted;
+  }
+*/
   while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
     hint = sen_str_get_prefix_order(_sen_sym_key(i->lexicon, *tp));
     if (hint == -1) { hint = *tp; }
@@ -980,6 +1050,10 @@ index_add(sen_index *i, const void *key, const char *value)
   sen_set_close(h);
   //  sen_log("add < (%x:%d:%d) %d", key, rid, *((int *)key), strlen(value));
   return rc;
+exit:
+  sen_set_close(h);
+  sen_lex_close(lex);
+  return sen_memory_exhausted;
 }
 
 inline static sen_rc
@@ -992,7 +1066,7 @@ index_del(sen_index *i, const void *key, const char *value)
   sen_id rid, tid, *tp;
   // sen_log("deleting(%s)", value);
   if (!(rid = sen_sym_at(i->keys, key))) {
-    sen_log("del : (%x) (invalid key)", key);
+    SEN_LOG(sen_log_error, "del : (%x) (invalid key)", key);
     return sen_invalid_argument;
   }
   // sen_log("del > (%x:%d:%d) %d", key, rid, *((int *)key), strlen(value));
@@ -1000,15 +1074,34 @@ index_del(sen_index *i, const void *key, const char *value)
     return sen_memory_exhausted;
   }
   h = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
+  if (!h) {
+    SEN_LOG(sen_log_alert, "sen_set_open on index_del failed !");
+    sen_lex_close(lex);
+    return sen_memory_exhausted;
+  }
   while (!lex->status) {
     if ((tid = sen_lex_next(lex))) {
       sen_set_get(h, &tid, (void **) &u);
-      if (!*u) { *u = sen_inv_updspec_open(rid, 0); }
+      if (!*u) {
+        if (!(*u = sen_inv_updspec_open(rid, 0))) {
+          SEN_LOG(sen_log_alert, "sen_inv_updspec_open on index_del failed !");
+          sen_set_close(h);
+          sen_lex_close(lex);
+          return sen_memory_exhausted;
+        }
+      }
     }
     // sen_log("index_del: tid=%d(%s)", tid, _sen_sym_key(i->lexicon, tid));
   }
   sen_lex_close(lex);
   c = sen_set_cursor_open(h);
+  SEN_ASSERT(c);
+/*
+  if (!c) {
+    // fix: free inv_updspec array *u
+    sen_set_close(h); return sen_memory_exhausted;
+  }
+*/
   while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
     if (*tp) {
       // sen_log("inv_delete > %d '%s'", *tp, _sen_sym_key(i->lexicon, *tp));
@@ -1025,18 +1118,23 @@ index_del(sen_index *i, const void *key, const char *value)
 
 sen_rc
 sen_index_upd(sen_index *i, const void *key,
-              const char *oldvalue, const char *newvalue)
+              const char *oldvalue, unsigned int oldvalue_len,
+              const char *newvalue, unsigned int newvalue_len)
 {
+  // TODO: use oldvalue_len/newvalue_len
   // if (!strcmp(oldvalue, newvalue)) { return sen_success; }
   sen_rc rc;
+  if (!i || !key) { SEN_LOG(sen_log_warning, "sen_index_upd: invalid argument"); return sen_invalid_argument; }
   if ((rc = sen_index_lock(i, -1))) {
-    sen_log("index lock failed");
+    SEN_LOG(sen_log_error, "sen_index_upd: index lock failed");
     return rc;
   }
   // sen_log("index_upd %p %x > '%s'->'%s'", i, key, oldvalue, newvalue);
   if (oldvalue && *oldvalue) {
-    rc = index_del(i, key, oldvalue); 
-    if (rc != sen_success) { goto exit; }
+    if ((rc = index_del(i, key, oldvalue))) {
+      SEN_LOG(sen_log_error, "index_del on sen_index_upd failed !");
+      goto exit;
+    }
   }
   if (newvalue && *newvalue) {
     rc = index_add(i, key, newvalue);
@@ -1055,7 +1153,9 @@ exit :
 sen_rc
 sen_index_del(sen_index *i, const void *key)
 {
-  sen_id rid = sen_sym_at(i->keys, key);
+  sen_id rid;
+  if (!i || !key) { SEN_LOG(sen_log_warning, "sen_index_del: invalid argument"); return sen_invalid_argument; }
+  rid = sen_sym_at(i->keys, key);
   if (!rid) { return sen_invalid_argument; }
   return sen_sym_pocket_set(i->keys, rid, DELETE_FLAG);
 }
@@ -1083,12 +1183,14 @@ sen_values_close(sen_values *v)
 }
 
 sen_rc
-sen_values_add(sen_values *v, const char *str, unsigned int weight)
+sen_values_add(sen_values *v, const char *str, unsigned int str_len, unsigned int weight)
 {
+  // TODO:use str_len
   sen_value *vp;
+  if (!v || !str) { SEN_LOG(sen_log_warning, "sen_values_add: invalid argument"); return sen_invalid_argument; }
   if (!(v->n_values & (INITIAL_VALUE_SIZE - 1))) {
     vp = SEN_REALLOC(v->values, sizeof(sen_value) * (v->n_values + INITIAL_VALUE_SIZE));
-    sen_log("expanded values to %d,%p", v->n_values + INITIAL_VALUE_SIZE, vp);
+    SEN_LOG(sen_log_debug, "expanded values to %d,%p", v->n_values + INITIAL_VALUE_SIZE, vp);
     if (!vp) { return sen_memory_exhausted; }
     v->values = vp;
   }
@@ -1112,8 +1214,12 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
   sen_set *old, *new;
   sen_id rid, tid, *tp;
   sen_inv_updspec **u, **un;
+  if (!i || !key) {
+    SEN_LOG(sen_log_warning, "sen_index_update: invalid argument");
+    return sen_invalid_argument;
+  }
   if ((rc = sen_index_lock(i, -1))) {
-    sen_log("index lock failed");
+    SEN_LOG(sen_log_crit, "sen_index_update: index lock failed");
     return rc;
   }
   if (newvalues) {
@@ -1122,13 +1228,32 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
       goto exit;
     }
     new = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
+    if (!new) {
+      SEN_LOG(sen_log_alert, "sen_set_open on sen_index_update failed !");
+      rc = sen_memory_exhausted;
+      goto exit;
+    }
     for (j = newvalues->n_values, v = newvalues->values; j; j--, v++) {
       if ((lex = sen_lex_open(i->lexicon, v->str, SEN_LEX_ADD|SEN_LEX_UPD))) {
         while (!lex->status) {
           if ((tid = sen_lex_next(lex))) {
             sen_set_get(new, &tid, (void **) &u);
-            if (!*u) { *u = sen_inv_updspec_open(rid, section); }
-            sen_inv_updspec_add(*u, lex->pos, v->weight);
+            if (!*u) {
+              if (!(*u = sen_inv_updspec_open(rid, section))) {
+                SEN_LOG(sen_log_alert, "sen_inv_updspec_open on sen_index_update failed!");
+                sen_lex_close(lex);
+                sen_set_close(new);
+                rc = sen_memory_exhausted;
+                goto exit;
+              }
+            }
+            if (sen_inv_updspec_add(*u, lex->pos, v->weight)) {
+              SEN_LOG(sen_log_alert, "sen_inv_updspec_add on sen_index_update failed!");
+              sen_lex_close(lex);
+              sen_set_close(new);
+              rc = sen_memory_exhausted;
+              goto exit;
+            }
           }
         }
         sen_lex_close(lex);
@@ -1147,13 +1272,35 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
   }
   if (oldvalues) {
     old = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
+    if (!old) {
+      SEN_LOG(sen_log_alert, "sen_set_open(old) on sen_index_update failed!");
+      if (new) { sen_set_close(new); }
+      rc = sen_memory_exhausted;
+      goto exit;
+    }
     for (j = oldvalues->n_values, v = oldvalues->values; j; j--, v++) {
       if ((lex = sen_lex_open(i->lexicon, v->str, SEN_LEX_UPD))) {
         while (!lex->status) {
           if ((tid = sen_lex_next(lex))) {
             sen_set_get(old, &tid, (void **) &u);
-            if (!*u) { *u = sen_inv_updspec_open(rid, section); }
-            sen_inv_updspec_add(*u, lex->pos, v->weight);
+            if (!*u) {
+              if (!(*u = sen_inv_updspec_open(rid, section))) {
+                SEN_LOG(sen_log_alert, "sen_inv_updspec_open on sen_index_update failed!");
+                sen_lex_close(lex);
+                if (new) { sen_set_close(new); };
+                sen_set_close(old);
+                rc = sen_memory_exhausted;
+                goto exit;
+              }
+            }
+            if (sen_inv_updspec_add(*u, lex->pos, v->weight)) {
+              SEN_LOG(sen_log_alert, "sen_inv_updspec_add on sen_index_update failed!");
+              sen_lex_close(lex);
+              if (new) { sen_set_close(new); };
+              sen_set_close(old);
+              rc = sen_memory_exhausted;
+              goto exit;
+            }
           }
         }
         sen_lex_close(lex);
@@ -1164,6 +1311,13 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
   }
   if (old) {
     c = sen_set_cursor_open(old);
+    SEN_ASSERT(c);
+/*
+    if (!c) {
+      // fixme: delete inv_updspec array *u
+      sen_set_close(old); if (new) { sen_set_close(new); } rc = sen_memory_exhausted; goto exit;
+    }
+*/
     while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
       if (new && (eh = sen_set_at(new, tp, (void **) &un))) {
         if (!sen_inv_updspec_cmp(*u, *un)) {
@@ -1180,6 +1334,13 @@ sen_index_update(sen_index *i, const void *key, unsigned int section,
   }
   if (new) {
     c = sen_set_cursor_open(new);
+    SEN_ASSERT(c);
+/*
+    if (!c) {
+      // fixme: delete inv_updspec array *u
+      sen_set_close(new); rc = sen_memory_exhausted; goto exit;
+    }
+*/
     while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
       hint = sen_str_get_prefix_order(_sen_sym_key(i->lexicon, *tp));
       if (hint == -1) { hint = *tp; }
@@ -1272,7 +1433,7 @@ sen_records_close(sen_records *r)
     sen_id *rid;
     recinfo *ri;
     if (!sen_set_element_info(r->records, r->curr_rec, (void **)&rid, (void **)&ri)) {
-      sen_log("curr_rec: %d:%d", *rid, ri->score);
+      SEN_LOG(sen_log_debug, "curr_rec: %d:%d", *rid, ri->score);
       // sen_log("sen_record_next=%d,%d", *rid, *(_sen_sym_key(r->keys, *rid)));
     }
   }
@@ -1284,7 +1445,7 @@ sen_records_close(sen_records *r)
 
 sen_rc
 sen_records_reopen(sen_records *r, sen_rec_unit record_unit,
-		   sen_rec_unit subrec_unit, unsigned int max_n_subrecs)
+                   sen_rec_unit subrec_unit, unsigned int max_n_subrecs)
 {
   int record_size = rec_unit_size(record_unit);
   int subrec_size = rec_unit_size(subrec_unit);
@@ -1373,8 +1534,8 @@ sen_records_next(sen_records *r, void *keybuf, int bufsize, int *score)
       if (r->record_unit != sen_rec_userdef) {
         return sen_sym_key(r->keys, *rid, keybuf, bufsize);
       } else {
-	if (r->record_size <= bufsize) { memcpy(keybuf, rid, r->record_size); }
-	return r->record_size;
+        if (r->record_size <= bufsize) { memcpy(keybuf, rid, r->record_size); }
+        return r->record_size;
       }
     }
   }
@@ -1563,14 +1724,24 @@ sen_records_group(sen_records *r, int limit, sen_group_optarg *optarg)
   if (funcp) {
     ssize = r->record_size;
     gkey = SEN_MALLOC(rsize ? rsize : 8192);
+    if (!gkey) {
+      SEN_LOG(sen_log_alert, "allocation for gkey failed !");
+      return sen_memory_exhausted;
+    }
   } else {
     if (r->record_size <= rsize) { return sen_invalid_argument; }
     ssize = r->record_size - rsize;
   }
   unit_size = SCORE_SIZE + ssize;
-  if (!(c = sen_set_cursor_open(r->records))) { return sen_memory_exhausted; }
+  if (!(c = sen_set_cursor_open(r->records))) {
+    SEN_LOG(sen_log_alert, "sen_set_cursor_open on sen_records_group failed !");
+    if (gkey) { SEN_FREE(gkey); }
+    return sen_memory_exhausted;
+  }
   if (!(g = sen_set_open(rsize, SCORE_SIZE + sizeof(int) + unit_size * limit, 0))) {
+    SEN_LOG(sen_log_alert, "sen_set_open on sen_records_group failed !");
     sen_set_cursor_close(c);
+    if (gkey) { SEN_FREE(gkey); }
     return sen_memory_exhausted;
   }
   while ((rh = sen_set_cursor_next(c, (void **)&key, (void **)&ri))) {
@@ -1979,6 +2150,11 @@ sen_index_similar_search(sen_index *i, const char *string,
   {
     sen_set_cursor *c = sen_set_cursor_open(h);
     int maxsize = sen_sym_size(i->keys) * sizeof(int);
+    if (!c) {
+      SEN_LOG(sen_log_alert, "sen_set_cursor_open on sen_index_similar_search failed !");
+      sen_set_close(h);
+      return sen_memory_exhausted;
+    }
     while (sen_set_cursor_next(c, (void **) &tp, (void **) &w1)) {
       *w1 += maxsize / (sen_inv_estimate_size(i->inv, *tp) + 1);
     }
@@ -1996,6 +2172,11 @@ sen_index_similar_search(sen_index *i, const char *string,
     sen_wv_mode wvm = sen_wv_none;
     sen_set_sort_optarg arg = {sen_sort_descending, NULL, (void *)sizeof(sen_id), NULL};
     sen_set_eh *eh, *sorted = sen_set_sort(h, limit, &arg);
+    if (!sorted) {
+      SEN_LOG(sen_log_alert, "sen_set_sort on sen_index_similar_search failed !");
+//      sen_set_close(h);
+//      return sen_memory_exhausted;
+    }
     rep = (r->record_unit == sen_rec_position || r->subrec_unit == sen_rec_position);
     if (optarg->func) {
       wvm = sen_wv_dynamic;
@@ -2004,8 +2185,8 @@ sen_index_similar_search(sen_index *i, const char *string,
     }
     for (j = 0, eh = sorted; j < limit; j++, eh++) {
       sen_set_element_info(h, eh, (void **) &tp, (void **) &w1);
-      if (!*tp || !(c = sen_inv_cursor_open(i->inv, *tp))) {
-        sen_log("cursor open failed (%d)", *tp);
+      if (!*tp || !(c = sen_inv_cursor_open(i->inv, *tp, rep))) {
+        SEN_LOG(sen_log_error, "cursor open failed (%d)", *tp);
         continue;
       }
       if (rep) {
@@ -2034,6 +2215,10 @@ sen_index_similar_search(sen_index *i, const char *string,
     recinfo *ri;
     sen_set_eh *eh;
     sen_set_cursor *c = sen_set_cursor_open(r->records);
+    if (!c) {
+      SEN_LOG(sen_log_alert, "sen_set_cursor_open on sen_index_similar_search failed !");
+      return sen_memory_exhausted;
+    }
     while ((eh = sen_set_cursor_next(c, NULL, (void **) &ri))) {
       if ((ri->n_subrecs & B31)) {
         ri->n_subrecs &= B30_00;
@@ -2052,7 +2237,7 @@ sen_index_similar_search(sen_index *i, const char *string,
 
 sen_rc
 sen_index_term_extract(sen_index *i, const char *string,
-		       sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
+                       sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
 {
   posinfo pi;
   sen_id tid;
@@ -2079,34 +2264,34 @@ sen_index_term_extract(sen_index *i, const char *string,
   for (p = string, position = 0; *p; p += skip, position += skip) {
     if ((tid = sen_sym_common_prefix_search(sym, p))) {
       if (policy == TERM_EXTRACT_EACH_POST) {
-	skip = sen_sym_key(sym, tid, keybuf, SEN_SYM_MAX_KEY_SIZE) - 1;
+        skip = sen_sym_key(sym, tid, keybuf, SEN_SYM_MAX_KEY_SIZE) - 1;
       } else {
-	skip = (int)sen_str_charlen(p, sym->encoding);
+        skip = (int)sen_str_charlen(p, sym->encoding);
       }
-      if (!(c = sen_inv_cursor_open(i->inv, tid))) {
-        sen_log("cursor open failed (%d)", tid);
+      if (!(c = sen_inv_cursor_open(i->inv, tid, rep))) {
+        SEN_LOG(sen_log_error, "cursor open failed (%d)", tid);
         continue;
       }
       if (rep) {
         while (!sen_inv_cursor_next(c)) {
           pos = c->post;
-	  while (!sen_inv_cursor_next_pos(c)) {
-	    res_add(r, (posinfo *) pos,
-		    get_weight(r, pos->rid, pos->sid, wvm, optarg), op);
+          while (!sen_inv_cursor_next_pos(c)) {
+            res_add(r, (posinfo *) pos,
+                    get_weight(r, pos->rid, pos->sid, wvm, optarg), op);
           }
         }
       } else {
-	while (!sen_inv_cursor_next(c)) {
-	  if (policy == TERM_EXTRACT_EACH_POST) {
-	    pi.rid = c->post->rid;
-	    pi.sid = position;
-	    res_add(r, &pi, position + 1, op);
-	  } else {
-	    pos = c->post;
-	    res_add(r, (posinfo *) pos,
-		    get_weight(r, pos->rid, pos->sid, wvm, optarg), op);
-	  }
-	}
+        while (!sen_inv_cursor_next(c)) {
+          if (policy == TERM_EXTRACT_EACH_POST) {
+            pi.rid = c->post->rid;
+            pi.sid = position;
+            res_add(r, &pi, position + 1, op);
+          } else {
+            pos = c->post;
+            res_add(r, (posinfo *) pos,
+                    get_weight(r, pos->rid, pos->sid, wvm, optarg), op);
+          }
+        }
       }
       sen_inv_cursor_close(c);
     } else {
@@ -2125,9 +2310,10 @@ sen_index_term_extract(sen_index *i, const char *string,
 }
 
 sen_rc
-sen_index_select(sen_index *i, const char *string,
+sen_index_select(sen_index *i, const char *string, unsigned int string_len,
                  sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
 {
+  // TODO: use string_len
   btr *bt = NULL;
   sen_rc rc = sen_success;
   int rep, orp, weight, max_interval = 0;
@@ -2169,7 +2355,7 @@ sen_index_select(sen_index *i, const char *string,
     sen_log("o=%d n=%d s=%d r=%d", ti->offset, ti->ntoken, ti->size, ti->rid);
   }
   */
-  sen_log("n=%d (%s)", n, string);
+  SEN_LOG(sen_log_info, "n=%d (%s)", n, string);
   if (n == 1 && (*tis)->cursors->n_entries == 1 && op == sen_sel_or
       && !r->records->n_entries
       && r->record_unit == sen_rec_document && !r->max_n_subrecs) {
@@ -2179,9 +2365,9 @@ sen_index_select(sen_index *i, const char *string,
       recinfo *ri;
       sen_inv_posting *p = c->post;
       if ((weight = get_weight(r, p->rid, p->sid, wvm, optarg))) {
-	SEN_SET_INT_ADD(r->records, p, ri);
-	ri->score = (p->tf + p->score) * weight;
-	ri->n_subrecs = 1;
+        SEN_SET_INT_ADD(r->records, p, ri);
+        ri->score = (p->tf + p->score) * weight;
+        ri->n_subrecs = 1;
       }
     } while (!sen_inv_cursor_next(c));
     goto exit;
@@ -2212,10 +2398,10 @@ sen_index_select(sen_index *i, const char *string,
     break; \
   } \
 }
-	if (n == 1 && !rep) {
-	  noccur = (*tis)->p->tf;
-	  tscore = (*tis)->p->score;
-	} else if (mode == sen_sel_near) {
+        if (n == 1 && !rep) {
+          noccur = (*tis)->p->tf;
+          tscore = (*tis)->p->score;
+        } else if (mode == sen_sel_near) {
           bt_zap(bt);
           for (tip = tis; tip < tie; tip++) {
             ti = *tip;
@@ -2248,9 +2434,9 @@ sen_index_select(sen_index *i, const char *string,
             ti = *tip;
             SKIP_OR_BREAK(pos);
             if (ti->pos == pos) {
-	      score += ti->p->score; count++;
+              score += ti->p->score; count++;
             } else {
-	      score = ti->p->score; count = 1; pos = ti->pos;
+              score = ti->p->score; count = 1; pos = ti->pos;
             }
             if (count == n) {
               if (rep) { pi.pos = pos; res_add(r, &pi, (score + 1) * weight, op); }
@@ -2274,14 +2460,19 @@ exit :
     recinfo *ri;
     sen_set_eh *eh;
     sen_set_cursor *c = sen_set_cursor_open(r->records);
-    while ((eh = sen_set_cursor_next(c, NULL, (void **) &ri))) {
-      if ((ri->n_subrecs & B31)) {
-        ri->n_subrecs &= B30_00;
-      } else {
-        sen_set_del(r->records, eh);
+    if (c) {
+      while ((eh = sen_set_cursor_next(c, NULL, (void **) &ri))) {
+        if ((ri->n_subrecs & B31)) {
+          ri->n_subrecs &= B30_00;
+        } else {
+          sen_set_del(r->records, eh);
+        }
       }
+      sen_set_cursor_close(c);
+    }
+    else {
+      SEN_LOG(sen_log_alert, "sen_set_cursor_open on sen_index_select failed !");
     }
-    sen_set_cursor_close(c);
   }
   sen_records_cursor_clear(r);
   bt_close(bt);
@@ -2290,7 +2481,7 @@ exit :
     uint32_t segno = SEN_INV_MAX_SEGMENT, nnref = 0;
     sen_io_mapinfo *info = i->inv->seg->maps;
     for (; segno; segno--, info++) { if (info->nref) { nnref++; } }
-    sen_log("nnref=%d", nnref);
+    SEN_LOG(sen_log_info, "nnref=%d", nnref);
   }
 #endif /* DEBUG */
   sen_inv_seg_expire(i->inv);
@@ -2298,25 +2489,38 @@ exit :
 }
 
 sen_records *
-sen_index_sel(sen_index *i, const char *string)
+sen_index_sel(sen_index *i, const char *string, unsigned int string_len)
 {
-  sen_log("sen_index_sel > (%s)", string);
+  SEN_LOG(sen_log_info, "sen_index_sel > (%s)", string);
   {
   sen_select_optarg arg = {sen_sel_exact, 0, 0, NULL, 0, NULL, NULL};
   sen_records *r = sen_records_open(sen_rec_document, sen_rec_none, 0);
-  sen_index_select(i, string, r, sen_sel_or, &arg);
-  sen_log("exact: %d", r->records->n_entries);
+  if (!r) { return NULL; }
+  if (sen_index_select(i, string, string_len, r, sen_sel_or, &arg)) {
+    SEN_LOG(sen_log_error, "sen_index_select on sen_index_sel(1) failed !");
+    SEN_FREE(r);
+    return NULL;
+  }
+  SEN_LOG(sen_log_info, "exact: %d", r->records->n_entries);
   if (r->records->n_entries <= sen_query_escalation_threshold) {
     arg.mode = sen_sel_unsplit;
-    sen_index_select(i, string, r, sen_sel_or, &arg);
-    sen_log("unsplit: %d", r->records->n_entries);
+    if (sen_index_select(i, string, string_len, r, sen_sel_or, &arg)) {
+      SEN_LOG(sen_log_error, "sen_index_select on sen_index_sel(2) failed !");
+      SEN_FREE(r);
+      return NULL;
+    }
+    SEN_LOG(sen_log_info, "unsplit: %d", r->records->n_entries);
   }
   if (r->records->n_entries <= sen_query_escalation_threshold) {
     arg.mode = sen_sel_partial;
-    sen_index_select(i, string, r, sen_sel_or, &arg);
-    sen_log("partial: %d", r->records->n_entries);
+    if (sen_index_select(i, string, string_len, r, sen_sel_or, &arg)) {
+      SEN_LOG(sen_log_error, "sen_index_select on sen_index_sel(3) failed !");
+      SEN_FREE(r);
+      return NULL;
+    }
+    SEN_LOG(sen_log_info, "partial: %d", r->records->n_entries);
   }
-  sen_log("hits=%d", r->records->n_entries);
+  SEN_LOG(sen_log_info, "hits=%d", r->records->n_entries);
   if (!r->records->n_entries) {
     sen_records_close(r);
     r = NULL;
@@ -2426,29 +2630,29 @@ sen_records_heap_next(sen_records_heap *h)
     {
       int n = 0, m = h->n_entries;
       if (m > 1) {
-	for (;;) {
-	  int n1 = n * 2 + 1;
-	  int n2 = n1 + 1;
-	  sen_records *r1 = n1 < m ? h->bins[n1] : NULL;
-	  sen_records *r2 = n2 < m ? h->bins[n2] : NULL;
-	  if (r1 && records_heap_cmp(h, r, r1)) {
-	    if (r2 && records_heap_cmp(h, r, r2) && records_heap_cmp(h, r1, r2)) {
-	      h->bins[n] = r2;
-	      n = n2;
-	    } else {
-	      h->bins[n] = r1;
-	      n = n1;
-	    }
-	  } else {
-	    if (r2 && records_heap_cmp(h, r, r2)) {
-	      h->bins[n] = r2;
-	      n = n2;
-	    } else {
-	      h->bins[n] = r;
-	      break;
-	    }
-	  }
-	}
+        for (;;) {
+          int n1 = n * 2 + 1;
+          int n2 = n1 + 1;
+          sen_records *r1 = n1 < m ? h->bins[n1] : NULL;
+          sen_records *r2 = n2 < m ? h->bins[n2] : NULL;
+          if (r1 && records_heap_cmp(h, r, r1)) {
+            if (r2 && records_heap_cmp(h, r, r2) && records_heap_cmp(h, r1, r2)) {
+              h->bins[n] = r2;
+              n = n2;
+            } else {
+              h->bins[n] = r1;
+              n = n1;
+            }
+          } else {
+            if (r2 && records_heap_cmp(h, r, r2)) {
+              h->bins[n] = r2;
+              n = n2;
+            } else {
+              h->bins[n] = r;
+              break;
+            }
+          }
+        }
       }
       h->curr++;
       return m;
