@@ -33,21 +33,13 @@
 #include "store.h"
 #endif /* _SEN_STORE_H */
 
-#ifdef	__cplusplus
+#ifdef  __cplusplus
 extern "C" {
 #endif
 
 #define SEN_INV_MAX_SEGMENT 0x4000
 
-struct sen_inv_header {
-  char idstr[16];
-  uint32_t initial_n_segments;
-  // todo: initial_n_segments should be uint16_t
-  // uint32_t total_chunk_size; todo: should be added when index format changed
-  uint16_t segments[SEN_INV_MAX_SEGMENT];
-  // todo: exchange segments and ainfo,binfo.
-  uint8_t chunks[1]; /* dummy */
-};
+struct sen_inv_header;
 
 struct _sen_inv_pos {
   struct _sen_inv_pos *next;
@@ -64,27 +56,15 @@ struct _sen_inv_updspec {
   sen_store_vnode *vnodes;
 };
 
-struct _sen_inv {
-  sen_io *seg;
-  sen_io *chunk;
-  sen_sym *lexicon;
-  struct sen_inv_header *header;
-  uint32_t total_chunk_size;
-  uint16_t ainfo[SEN_INV_MAX_SEGMENT];
-  uint16_t binfo[SEN_INV_MAX_SEGMENT];
-  uint16_t amax;
-  uint16_t bmax;
-};
-
 typedef struct _sen_inv_updspec sen_inv_updspec;
 
 sen_inv * sen_inv_create(const char *path, sen_sym *lexicon, 
-			 int flags, uint32_t initial_n_segments);
+                         int flags, uint32_t initial_n_segments);
 sen_inv * sen_inv_open(const char *path, sen_sym *lexicon);
 sen_rc sen_inv_close(sen_inv *inv);
 sen_rc sen_inv_info(sen_inv *inv, off_t *seg_size, off_t *chunk_size);
 sen_rc sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h,
-		      int hint);
+                      int hint);
 sen_rc sen_inv_delete(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h);
 uint32_t sen_inv_initial_n_segments(sen_inv *inv);
 
@@ -107,10 +87,12 @@ typedef struct {
 } sen_inv_posting;
 
 typedef struct {
+  sen_inv *inv;
   sen_inv_posting pc;
   sen_inv_posting pb;
   sen_inv_posting *post;
   uint8_t *cp;
+  uint8_t *cpp;
   uint8_t *cpe;
   uint8_t *bp;
   sen_io_win iw;
@@ -118,7 +100,8 @@ typedef struct {
   uint16_t stat;
   uint16_t nextb;
   uint32_t buffer_pos;
-  sen_inv *inv;
+  int with_pos;
+  int flags;
 } sen_inv_cursor;
 
 #define SEN_INV_CURSOR_CMP(c1,c2) \
@@ -128,7 +111,7 @@ typedef struct {
      (((c1)->post->sid == (c2)->post->sid) && \
       ((c1)->post->pos > (c2)->post->pos)))))
 
-sen_inv_cursor *sen_inv_cursor_open(sen_inv *inv, uint32_t key);
+sen_inv_cursor *sen_inv_cursor_open(sen_inv *inv, uint32_t key, int with_pos);
 sen_inv_cursor *sen_inv_cursor_openv1(sen_inv *inv, uint32_t key);
 sen_rc sen_inv_cursor_openv2(sen_inv_cursor **cursors, int ncursors);
 sen_rc sen_inv_cursor_next(sen_inv_cursor *c);
