@@ -29,7 +29,7 @@
 #include "io.h"
 #include "cache.h"
 
-#define SEN_IO_IDSTR "SENNA:IO:00.001"
+#define SEN_IO_IDSTR "SENNA:IO:01.000"
 
 /* VA hack */
 /* max aio request (/proc/sys/fs/aio-max-nr) */
@@ -41,12 +41,7 @@
 
 static unsigned int pagesize = 0;
 
-typedef struct {
-  void *map;
-  void **handle;
-} mapinfo;
-
-typedef struct {
+typedef struct _sen_io_fileinfo {
 #ifdef WIN32
   HANDLE fh;
   HANDLE fmo;
@@ -58,25 +53,21 @@ typedef struct {
 #endif /* WIN32 */
 } fileinfo;
 
-typedef struct {
+typedef struct _sen_io_header08 {
   char idstr[16];
   uint32_t header_size;
   uint32_t segment_size;
   uint32_t max_segment;
-} io_header;
+} io_header08;
 
-struct _sen_io {
-  char path[PATH_MAX];
-  io_header *header;
-  byte *user_header;
-  mapinfo *maps;
-  uint32_t base;
-  uint32_t base_seg;
-  sen_io_mode mode;
-  uint32_t cache_size;
-  fileinfo *fis;
-  sen_mutex map_mutex;
-};
+typedef struct _sen_io_header {
+  char idstr[16];
+  uint32_t header_size;
+  uint32_t segment_size;
+  uint32_t max_segment;
+  uint64_t curr_size;
+  uint32_t reserved[7];
+} io_header;
 
 inline static unsigned int sen_get_pagesize(void);
 inline static sen_rc sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize);
@@ -91,16 +82,17 @@ inline static sen_rc sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t off
 
 sen_io *
 sen_io_create(const char *path, uint32_t header_size, uint32_t segment_size,
-	      uint32_t max_segment, sen_io_mode mode, uint32_t cache_size)
+              uint32_t max_segment, sen_io_mode mode, uint32_t cache_size)
 {
   sen_io *io;
   fileinfo *fis;
   unsigned int b, max_nfiles;
-  uint32_t bs;
+  uint32_t bs, total_header_size;
   io_header *header;
   if (!path || !*path || (strlen(path) > PATH_MAX - 4)) { return NULL; }
   if (!pagesize) { pagesize = sen_get_pagesize(); }
-  for (b = pagesize; b < sizeof(io_header) + header_size; b += pagesize);
+  total_header_size = sizeof(io_header) + max_segment * sizeof(uint32_t) + header_size;
+  for (b = pagesize; b < total_header_size; b += pagesize);
   bs = (b + segment_size - 1) / segment_size;
   max_nfiles = (unsigned int)(
     ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1) 
@@ -115,20 +107,22 @@ sen_io_create(const char *path, uint32_t header_size, uint32_t segment_size,
       memcpy(header->idstr, SEN_IO_IDSTR, 16);
       sen_msync(header, b);
       if ((io = SEN_MALLOC(sizeof(sen_io)))) {
-	if ((io->maps = SEN_MALLOC(sizeof(mapinfo) * max_segment))) {
-	  memset(io->maps, 0, sizeof(mapinfo) * max_segment);
-	  strncpy(io->path, path, PATH_MAX);
-	  io->header = header;
-	  io->user_header = ((byte *) header) + sizeof(io_header);
-	  io->base = b;
-	  io->base_seg = bs;
-	  io->mode = mode;
-	  io->cache_size = cache_size;
-	  io->fis = fis;
-	  MUTEX_INIT(io->map_mutex);
-	  return io;
-	}
-	SEN_FREE(io);
+        if ((io->maps = SEN_MALLOC(sizeof(sen_io_mapinfo) * max_segment))) {
+          memset(io->maps, 0, sizeof(sen_io_mapinfo) * max_segment);
+          strncpy(io->path, path, PATH_MAX);
+          io->header = header;
+          io->nrefs = (uint32_t *)(((byte *) header) + sizeof(io_header));
+          io->user_header = ((byte *) io->nrefs) + max_segment * sizeof(uint32_t);
+          io->base = b;
+          io->base_seg = bs;
+          io->mode = mode;
+          io->cache_size = cache_size;
+          io->fis = fis;
+          io->nmaps = 0;
+          io->count = 0;
+          return io;
+        }
+        SEN_FREE(io);
       }
       sen_munmap(header, b);
     }
@@ -144,9 +138,9 @@ sen_io_open(const char *path, sen_io_mode mode, uint32_t cache_size)
   sen_io *io;
   struct stat s;
   fileinfo *fis;
-  io_header *header;
   unsigned int b, max_nfiles;
-  uint32_t header_size = 0, segment_size = 0, max_segment = 0, bs;
+  uint32_t total_header_size;
+  uint32_t header_size = 0, segment_size = 0, max_segment = 0, bs, v08p = 0;
   if (!path || !*path || (strlen(path) > PATH_MAX - 4)) { return NULL; }
   if (!pagesize) { pagesize = sen_get_pagesize(); }
   {
@@ -155,18 +149,24 @@ sen_io_open(const char *path, sen_io_mode mode, uint32_t cache_size)
     if (fd == -1) { return NULL; }
     if (fstat(fd, &s) != -1 && s.st_size >= sizeof(io_header)) {
       if (read(fd, &h, sizeof(io_header)) == sizeof(io_header)) {
-	/* if (!memcmp(header->idstr, SEN_IO_IDSTR, 16)) */
-	{
-	  header_size = h.header_size;
-	  segment_size = h.segment_size;
-	  max_segment = h.max_segment;
-	}
+        if (memcmp(h.idstr, SEN_IO_IDSTR, 16)) {
+          /* regard as v0.8 format */
+          v08p = 1;
+        }
+        header_size = h.header_size;
+        segment_size = h.segment_size;
+        max_segment = h.max_segment;
       }
     }
     close(fd);
     if (!segment_size) { return NULL; }
   }
-  for (b = pagesize; b < sizeof(io_header) + header_size; b += pagesize);
+  if (v08p) {
+    total_header_size = sizeof(io_header) + header_size;
+  } else {
+    total_header_size = sizeof(io_header) + header_size + max_segment * sizeof(uint32_t);
+  }
+  for (b = pagesize; b < total_header_size; b += pagesize);
   bs = (b + segment_size - 1) / segment_size;
   max_nfiles = (unsigned int)(
     ((uint64_t)segment_size * (max_segment + bs) + SEN_IO_FILE_SIZE - 1) 
@@ -174,22 +174,32 @@ sen_io_open(const char *path, sen_io_mode mode, uint32_t cache_size)
   if (!(fis = SEN_MALLOC(sizeof(fileinfo) * max_nfiles))) { return NULL; }
   sen_fileinfo_init(fis, max_nfiles);
   if (!sen_open(fis, path, O_RDWR, SEN_IO_FILE_SIZE)) {
+    io_header *header;
     if ((header = sen_mmap(fis, 0, b))) {
       if ((io = SEN_MALLOC(sizeof(sen_io)))) {
-	if ((io->maps = SEN_MALLOC(sizeof(mapinfo) * max_segment))) {
-	  memset(io->maps, 0, sizeof(mapinfo) * max_segment);
-	  strncpy(io->path, path, PATH_MAX);
-	  io->header = header;
-	  io->user_header = ((byte *) header) + sizeof(io_header);
-	  io->base = b;
-	  io->base_seg = bs;
-	  io->mode = mode;
-	  io->cache_size = cache_size;
-	  io->fis = fis;
-	  MUTEX_INIT(io->map_mutex);
-	  return io;
-	}
-	SEN_FREE(io);
+        if ((io->maps = SEN_CALLOC(sizeof(sen_io_mapinfo) * max_segment))) {
+          strncpy(io->path, path, PATH_MAX);
+          io->header = header;
+          if (v08p) {
+            io->nrefs = SEN_CALLOC(sizeof(uint32_t) * max_segment);
+            io->user_header = ((byte *) header) + sizeof(io_header08);
+          } else {
+            io->nrefs = (uint32_t *)(((byte *) header) + sizeof(io_header));
+            io->user_header = ((byte *) io->nrefs) + max_segment * sizeof(uint32_t);
+          }
+          if (io->nrefs) {
+            io->base = b;
+            io->base_seg = bs;
+            io->mode = mode;
+            io->cache_size = cache_size;
+            io->fis = fis;
+            io->nmaps = 0;
+            io->count = 0;
+            return io;
+          }
+          SEN_FREE(io->maps);
+        }
+        SEN_FREE(io);
       }
       sen_munmap(header, b);
     }
@@ -203,7 +213,7 @@ sen_rc
 sen_io_close(sen_io *io)
 {
   int i;
-  mapinfo *mi;
+  sen_io_mapinfo *mi;
   fileinfo *fi;
   uint32_t bs = io->base_seg;
   uint32_t max_segment = io->header->max_segment;
@@ -213,8 +223,8 @@ sen_io_close(sen_io *io)
     / SEN_IO_FILE_SIZE);
   for (mi = io->maps, i = max_segment; i; mi++, i--) {
     if (mi->map) {
+      /* if (atomic_read(mi->nref)) { return STILL_IN_USE ; } */
       sen_munmap(mi->map, segment_size);
-      if (mi->handle) { *mi->handle = NULL; }
     }
   }
   sen_munmap(io->header, io->base);
@@ -316,10 +326,10 @@ sen_io_rename(const char *old_name, const char *new_name)
     for (fno = 1; ; fno++) {
       gen_pathname(old_name, old_buffer, fno);
       if (!stat(old_buffer, &s)) {
-	gen_pathname(new_name, new_buffer, fno);
-	rename(old_buffer, new_buffer);
+        gen_pathname(new_name, new_buffer, fno);
+        rename(old_buffer, new_buffer);
       } else {
-	break;
+        break;
       }
     }
     return sen_success;
@@ -329,7 +339,7 @@ sen_io_rename(const char *old_name, const char *new_name)
 
 void *
 sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
-	       uint32_t offset, uint32_t size, sen_io_rw_mode mode)
+               uint32_t offset, uint32_t size, sen_io_rw_mode mode)
 {
   byte *p;
   off_t pos;
@@ -356,16 +366,16 @@ sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
     {
       fileinfo *fi = &io->fis[fno];
       if (!sen_opened(fi)) {
-	char path[PATH_MAX];
-	gen_pathname(io->path, path, fno);
-	if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
-	  return NULL;
-	}
+        char path[PATH_MAX];
+        gen_pathname(io->path, path, fno);
+        if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
+          return NULL;
+        }
       }
       if (!(p = SEN_MALLOC(size))) { return NULL; }
       if (sen_pread(fi, p, size, pos)) {
-	SEN_FREE(p);
-	return NULL;
+        SEN_FREE(p);
+        return NULL;
       }
       iw->addr = p;
     }
@@ -375,17 +385,17 @@ sen_io_win_map(sen_io *io, sen_io_win *iw, uint32_t segment,
     if (nseg > 0) {
       fileinfo *fi = &io->fis[fno];
       if (!sen_opened(fi)) {
-	char path[PATH_MAX];
-	gen_pathname(io->path, path, fno);
-	if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
-	  return NULL;
-	}
+        char path[PATH_MAX];
+        gen_pathname(io->path, path, fno);
+        if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
+          return NULL;
+        }
       }
       if (!(p = sen_mmap(fi, pos, nseg * segment_size))) {
-	return NULL;
+        return NULL;
       }
     } else {
-      if (!(p = sen_io_seg_map(io, segment))) { return NULL; }
+      if (!(p = sen_io_seg_ref(io, segment))) { return NULL; }
     }
     iw->addr = p + offset;
     break;
@@ -440,19 +450,19 @@ retry:
             segment += offset / segment_size;
             offset = offset % segment_size;
         }
-	nseg = (offset + size + segment_size - 1) / segment_size;
-	bseg = segment + io->base_seg;
-	fno = bseg / segments_per_file;
-	base = fno ? 0 : io->base - io->base_seg * segment_size;
-	pos = (bseg % segments_per_file) * segment_size + offset + base;
+        nseg = (offset + size + segment_size - 1) / segment_size;
+        bseg = segment + io->base_seg;
+        fno = bseg / segments_per_file;
+        base = fno ? 0 : io->base - io->base_seg * segment_size;
+        pos = (bseg % segments_per_file) * segment_size + offset + base;
         if (!size || !io || segment + nseg > io->header->max_segment ||
             fno != (bseg + nseg - 1) / segments_per_file) {
-	  return sen_other_error;
+          return sen_other_error;
         }
         fi = &io->fis[fno];
         if (!sen_opened(fi)) {
             char path[PATH_MAX];
-	    gen_pathname(io->path, path, fno);
+            gen_pathname(io->path, path, fno);
             if (sen_open(fi, path, O_RDWR|O_CREAT|O_DIRECT, SEN_IO_FILE_SIZE)) {
                 return sen_other_error;
             }
@@ -589,12 +599,12 @@ sen_io_win_unmap(sen_io_win *iw)
     if (nseg > 0) {
       sen_munmap(((byte *)iw->addr) - iw->offset, nseg * segment_size);
     } else {
-      rc = sen_io_seg_unmap(io, iw->segment);
+      rc = sen_io_seg_unref(io, iw->segment);
 #ifdef USE_AIO
       if (sen_aio_enabled) {
-	int fno = (iw->segment + io->base_seg) / segments_per_file;
-	fileinfo *fi = &io->fis[fno];
-	sen_cache_mark_invalid(fi->dev, fi->inode, iw->pos, iw->size);
+        int fno = (iw->segment + io->base_seg) / segments_per_file;
+        fileinfo *fi = &io->fis[fno];
+        sen_cache_mark_invalid(fi->dev, fi->inode, iw->pos, iw->size);
       }
 #endif /* USE_AIO */
     }
@@ -605,19 +615,19 @@ sen_io_win_unmap(sen_io_win *iw)
       int fno = (iw->segment + io->base_seg) / segments_per_file;
       fileinfo *fi = &io->fis[fno];
       if (!sen_opened(fi)) {
-	char path[PATH_MAX];
-	gen_pathname(io->path, path, fno);
-	rc = sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE);
+        char path[PATH_MAX];
+        gen_pathname(io->path, path, fno);
+        rc = sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE);
       }
       if (!rc) {
-	if (!(rc = sen_pwrite(fi, iw->addr, iw->size, iw->pos))) {
-	  SEN_FREE(iw->addr);
-	  iw->addr = NULL;
-	}
+        if (!(rc = sen_pwrite(fi, iw->addr, iw->size, iw->pos))) {
+          SEN_FREE(iw->addr);
+          iw->addr = NULL;
+        }
 #ifdef USE_AIO
         if (sen_aio_enabled) {
-	  sen_cache_mark_invalid(fi->dev, fi->inode, iw->pos, iw->size);
-	}
+          sen_cache_mark_invalid(fi->dev, fi->inode, iw->pos, iw->size);
+        }
 #endif /* USE_AIO */
       }
     }
@@ -628,99 +638,133 @@ sen_io_win_unmap(sen_io_win *iw)
   return rc;
 }
 
-void * 
-sen_io_seg_map(sen_io *io, int32_t segno)
-{
-  void *map;
-  mapinfo *info;
-  uint32_t segment_size = io->header->segment_size;
-  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;
+#define SEG_MAP(io,segno,info)\
+{\
+  uint32_t segment_size = io->header->segment_size;\
+  uint32_t segments_per_file = SEN_IO_FILE_SIZE / segment_size;\
+  uint32_t bseg = segno + io->base_seg;\
+  uint32_t fno = bseg / segments_per_file;\
+  uint32_t base = fno ? 0 : io->base - io->base_seg * segment_size;\
+  off_t pos = (bseg % segments_per_file) * segment_size + base;\
+  fileinfo *fi = &io->fis[fno];\
+  if (!sen_opened(fi)) {\
+    char path[PATH_MAX];\
+    gen_pathname(io->path, path, fno);\
+    if (!sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {\
+      if ((info->map = sen_mmap(fi, pos, segment_size))) {\
+        uint32_t nmaps;\
+        SEN_ATOMIC_ADD_EX(&io->nmaps, 1, nmaps);\
+      }\
+    }\
+  } else {\
+    if ((info->map = sen_mmap(fi, pos, segment_size))) {\
+      uint32_t nmaps;\
+      SEN_ATOMIC_ADD_EX(&io->nmaps, 1, nmaps);\
+    }\
+  }\
+}
 
-  if (0 > segno || (uint32_t)segno >= io->header->max_segment) { return NULL; }
+void
+sen_io_seg_map_(sen_io *io, uint32_t segno, sen_io_mapinfo *info)
+{
+  SEG_MAP(io, segno, info);
+}
+
+void * 
+sen_io_seg_ref(sen_io *io, uint32_t segno)
+{
+  uint32_t retry, *pnref;
+  sen_io_mapinfo *info;
+  if (segno >= io->header->max_segment) { return NULL; }
   info = &io->maps[segno];
-  if (!(map = info->map)) {
-    int bseg = segno + io->base_seg;
-    int fno = bseg / segments_per_file;
-    int base = fno ? 0 : io->base - io->base_seg * segment_size;
-    fileinfo *fi = &io->fis[fno];
-    off_t pos = (bseg % segments_per_file) * segment_size + base;
-    if (!sen_opened(fi)) {
-      char path[PATH_MAX];
-      gen_pathname(io->path, path, fno);
-      if (sen_open(fi, path, O_RDWR|O_CREAT, SEN_IO_FILE_SIZE)) {
-	return NULL;
+  /* pnref = &io->nrefs[segno]; */
+  pnref = &info->nref;
+  for (retry = 0;; retry++) {
+    uint32_t nref;
+    SEN_ATOMIC_ADD_EX(pnref, 1, nref);
+    if (nref >= SEN_IO_MAX_REF) {
+      SEN_ATOMIC_ADD_EX(pnref, -1, nref);
+      if (retry >= SEN_IO_MAX_RETRY) {
+        SEN_LOG(sen_log_crit, "deadlock detected! in sen_io_seg_ref(%p, %u)", io, segno);
+        return NULL;
+      }
+      usleep(1);
+      continue;
+    }
+    if (!info->map) {
+      if (nref) {
+        SEN_ATOMIC_ADD_EX(pnref, -1, nref);
+        if (retry >= SEN_IO_MAX_RETRY) {
+          SEN_LOG(sen_log_crit, "deadlock detected!! in sen_io_seg_ref(%p, %u)", io, segno);
+          return NULL;
+        }
+        usleep(1);
+        continue;
+      } else {
+        SEG_MAP(io, segno, info);
       }
     }
-    MUTEX_LOCK(io->map_mutex);
-    if (!(map = info->map)) {
-      if ((map = sen_mmap(fi, pos, segment_size))) {
-	info->map = map; // verbose
-      }
-    }
-    MUTEX_UNLOCK(io->map_mutex);
-    // sen_log("seg_map segno=%d map=%p", segno, map);
+    break;
   }
-  return map;
+  info->count = io->count++;
+  return info->map;
 }
 
 sen_rc
-sen_io_seg_unmap(sen_io *io, int32_t segno)
+sen_io_seg_unref(sen_io *io, uint32_t segno)
 {
-  void *map;
-  mapinfo *info;
-  if (0 > segno || (uint32_t)segno >= io->header->max_segment) { return sen_invalid_argument; }
-  info = &io->maps[segno];
-  // MUTEX_LOCK(io->map_mutex);
-  if ((map = info->map)) {
-    // todo:
-    // enqueue(io->garbases, segno);
-    //info->map = NULL;
-    if (info->handle) { *(info->handle) = NULL; }
-  }
-  // MUTEX_UNLOCK(io->map_mutex);
+  uint32_t nref;
+  uint32_t *pnref;
+  if (segno >= io->header->max_segment) { return sen_invalid_argument; }
+  pnref = &io->maps[segno].nref;
+  /* pnref = &io->nrefs[segno]; */
+  SEN_ATOMIC_ADD_EX(pnref, -1, nref);
   return sen_success;
 }
 
 sen_rc
-sen_io_seg_bind(sen_io *io, sen_io_seginfo *si)
+sen_io_seg_expire(sen_io *io, uint32_t segno)
 {
-  if ((si->addr = sen_io_seg_map(io, si->segno))) {
-    io->maps[si->segno].handle = &si->addr;
-    return sen_success;
-  } else {
-    return sen_memory_exhausted;
+  uint32_t retry, *pnref;
+  sen_io_mapinfo *info;
+  if (segno >= io->header->max_segment) { return sen_invalid_argument; }
+  info = &io->maps[segno];
+  /* pnref = &io->nrefs[segno]; */
+  pnref = &info->nref;
+  for (retry = 0;; retry++) {
+    uint32_t nref;
+    SEN_ATOMIC_ADD_EX(pnref, 1, nref);
+    if (nref) {
+      SEN_ATOMIC_ADD_EX(pnref, -1, nref);
+      if (retry >= SEN_IO_MAX_RETRY) {
+        SEN_LOG(sen_log_crit, "deadlock detected! in sen_io_seg_expire(%p, %u)", io, segno);
+        return sen_other_error;
+      }
+      usleep(1);
+      continue;
+    } else {
+      SEN_ATOMIC_ADD_EX(pnref, SEN_IO_MAX_REF, nref);
+      if (nref > 1) {
+        SEN_ATOMIC_ADD_EX(pnref, -(SEN_IO_MAX_REF + 1), nref);
+        if (retry >= SEN_IO_MAX_RETRY) {
+          SEN_LOG(sen_log_crit, "deadlock detected!! in sen_io_seg_expire(%p, %u)", io, segno);
+          return sen_other_error;
+        }
+        usleep(1);
+        continue;
+      } else {
+        uint32_t nmaps;
+        sen_munmap(info->map, io->header->segment_size);
+        info->map = NULL;
+        SEN_ATOMIC_ADD_EX(pnref, -(SEN_IO_MAX_REF + 1), nref);
+        SEN_ATOMIC_ADD_EX(&io->nmaps, -1, nmaps);
+        break;
+      }
+    }
   }
+  return sen_success;
 }
 
-sen_rc
-sen_io_seg_unbind(sen_io *io, sen_io_seginfo *si)
-{
-  sen_rc rc;
-  if (io->maps[si->segno].handle != &si->addr) {
-    sen_log("invalid unbind request");
-    return sen_invalid_argument;
-  }
-  rc = sen_io_seg_unmap(io, si->segno);
-  si->segno = -1;
-  si->addr = NULL;
-  return rc;
-}
-
-// todo: sen_io_gc()
-
-/*
-enque(io)
-{
-  pickup unused segments using mincore;
-  enqueue;
-}
-
-unmap(io)
-{
-  unmap segments in queue and passed enough time;
-}
-*/
-
 /** mmap abstraction **/
 
 static size_t mmap_size = 0;
@@ -741,15 +785,15 @@ sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
   char *pd, *ps, buf[PATH_MAX];
   /* may be wrong if flags is just only O_RDWR */
   fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
-		      FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
-		      OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
+                      FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
+                      OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
   if (fi->fh == INVALID_HANDLE_VALUE) { return sen_file_operation_error; }
   if ((flags & O_TRUNC)) {
     CloseHandle(fi->fh);
     /* unable to assign OPEN_ALWAYS and TRUNCATE_EXISTING at once */
     fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
-			FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
-			TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
+                        FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
+                        TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
     if (fi->fh == INVALID_HANDLE_VALUE) { return sen_file_operation_error; }
   }
   for (pd = buf, ps = path; *ps && pd < buf + PATH_MAX - 1; ps++, pd++) {
@@ -775,11 +819,11 @@ sen_open(fileinfo *fi, const char *path, int flags, size_t maxsize)
       MUTEX_INIT(fi->mutex);
       return sen_success;
     } else {
-      sen_log("fmo object already exists! handle=%d", fi->fh);
+      SEN_LOG(sen_log_error, "fmo object already exists! handle=%d", fi->fh);
       CloseHandle(fi->fmo);
     }
   } else {
-    sen_log("failed to get FileMappingObject #%d", GetLastError());
+    SEN_LOG(sen_log_alert, "failed to get FileMappingObject #%d", GetLastError());
   }
   CloseHandle(fi->fh);
   return sen_file_operation_error;
@@ -839,7 +883,7 @@ sen_mmap(fileinfo *fi, off_t offset, size_t length)
   if (res) {
     mmap_size += length;
   } else {
-    sen_log("MapViewOfFile failed #%d", GetLastError());
+    SEN_LOG(sen_log_alert, "MapViewOfFile failed #%d", GetLastError());
   }
   return res;
 }
@@ -858,7 +902,7 @@ sen_munmap(void *start, size_t length)
     mmap_size -= length;
     return 0;
   } else {
-    sen_log("munmap(%p,%d) failed <%d>", start, length, mmap_size);
+    SEN_LOG(sen_log_alert, "munmap(%p,%d) failed <%d>", start, length, mmap_size);
     return -1;
   }
 }
@@ -872,13 +916,13 @@ sen_pread(fileinfo *fi, void *buf, size_t count, off_t offset)
   r = SetFilePointer(fi->fh, offset, NULL, FILE_BEGIN);
   if (r == INVALID_SET_FILE_POINTER) {
     rc = sen_file_operation_error;
-    sen_log("SetFilePointer error(%d)", GetLastError());
+    SEN_LOG(sen_log_alert, "SetFilePointer error(%d)", GetLastError());
   } else {
     if (!ReadFile(fi->fh, buf, (DWORD)count, &len, NULL)) {
       rc = sen_file_operation_error;
-      sen_log("ReadFile error(%d)", GetLastError());
+      SEN_LOG(sen_log_alert, "ReadFile error(%d)", GetLastError());
     } else if (len != count) {
-      sen_log("ReadFile %d != %d", count, len);
+      SEN_LOG(sen_log_alert, "ReadFile %d != %d", count, len);
       rc = sen_file_operation_error;
     }
   }
@@ -895,13 +939,13 @@ sen_pwrite(fileinfo *fi, void *buf, size_t count, off_t offset)
   r = SetFilePointer(fi->fh, offset, NULL, FILE_BEGIN);
   if (r == INVALID_SET_FILE_POINTER) {
     rc = sen_file_operation_error;
-    sen_log("SetFilePointer error(%d)", GetLastError());
+    SEN_LOG(sen_log_alert, "SetFilePointer error(%d)", GetLastError());
   } else {
     if (!WriteFile(fi->fh, buf, (DWORD)count, &len, NULL)) {
-      sen_log("WriteFile error(%d)", GetLastError());
+      SEN_LOG(sen_log_alert, "WriteFile error(%d)", GetLastError());
       rc = sen_file_operation_error;
     } else if (len != count) {
-      sen_log("WriteFile %d != %d", count, len);
+      SEN_LOG(sen_log_alert, "WriteFile %d != %d", count, len);
       rc = sen_file_operation_error;
     }
   }
@@ -960,12 +1004,12 @@ sen_mmap(fileinfo *fi, off_t offset, size_t length)
   off_t tail = offset + length;
   if ((fstat(fi->fd, &s) == -1) ||
       (s.st_size < tail && ftruncate(fi->fd, tail) == -1)) {
-    sen_log("fstat or ftruncate failed %d", fi->fd);
+    SEN_LOG(sen_log_alert, "fstat or ftruncate failed %d", fi->fd);
     return NULL;
   }
   res = mmap(NULL, length, PROT_READ|PROT_WRITE, MAP_SHARED, fi->fd, offset);
   if (MAP_FAILED == res) {
-    sen_log("mmap(%d,%d,%d)=%p <%d>", length, fi->fd, offset, res, mmap_size);
+    SEN_LOG(sen_log_alert, "mmap(%z,%d,%d)=%s <%z>", length, fi->fd, offset, strerror(errno), mmap_size);
     return NULL;
   }
   mmap_size += length;
@@ -984,7 +1028,7 @@ sen_munmap(void *start, size_t length)
   int res;
   res = munmap(start, length);
   if (res) {
-    sen_log("munmap(%p,%d) failed <%d>", start, length, mmap_size);
+    SEN_LOG(sen_log_alert, "munmap(%p,%z) failed <%z>", start, length, mmap_size);
   } else {
     mmap_size -= length;
   }
