@@ -4,12 +4,12 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
@@ -160,8 +160,8 @@ get_word(sen_query *q, int *prefixp)
   const char *start = q->cur, *end;
   for (end = q->cur;; ) {
     if (!*end ||
-	sen_isspace(end, q->encoding) ||
-	*end == SEN_QUERY_PARENR) {
+        sen_isspace(end, q->encoding) ||
+        *end == SEN_QUERY_PARENR) {
       q->cur = end;
       break;
     }
@@ -194,7 +194,7 @@ get_op(sen_query *q, int *mode, int *option)
     start = ++end;
     while (isdigit(*end)) { end++; }
     if ((len = end - start)) {
-      buf = SEN_MALLOC(len + 1);
+      if (!(buf = SEN_MALLOC(len + 1))) { SEN_LOG(sen_log_alert, "get_op(S) malloc fail"); return; }
       memcpy(buf, start, len);
       buf[len] = '\0';
       *option = atoi(buf);
@@ -209,7 +209,7 @@ get_op(sen_query *q, int *mode, int *option)
     start = ++end;
     while (isdigit(*end)) { end++; }
     if ((len = end - start)) {
-      buf = SEN_MALLOC(len + 1);
+      if (!(buf = SEN_MALLOC(len + 1))) { SEN_LOG(sen_log_alert, "get_op(N) malloc fail"); return; }
       memcpy(buf, start, len);
       buf[len] = '\0';
       *option = atoi(buf);
@@ -224,7 +224,7 @@ get_op(sen_query *q, int *mode, int *option)
     start = ++end;
     while (isdigit(*end)) { end++; }
     if ((len = end - start)) {
-      buf = SEN_MALLOC(len + 1);
+      if (!(buf = SEN_MALLOC(len + 1))) { SEN_LOG(sen_log_alert, "get_op(T) malloc fail"); return; }
       memcpy(buf, start, len);
       buf[len] = '\0';
       *option = atoi(buf);
@@ -289,14 +289,13 @@ get_token(sen_query *q)
       break;
     default :
       if ((token = get_word(q, &prefixp)) &&
-	  token->u.token.start[0] == 'O' &&
-	  token->u.token.start[1] == 'R' &&
-	  token->u.token.len == 2) {
-	op = sen_sel_or;
-	cons_del(q);
-	token = NULL;
+          token->u.token.start[0] == 'O' &&
+          token->u.token.start[1] == 'R' &&
+          token->u.token.len == 2) {
+        op = sen_sel_or;
+        cons_del(q);
+        token = NULL;
       }
-      
       break;
     }
   } while (!token);
@@ -343,47 +342,47 @@ get_pragma(sen_query *q)
       start = ++end;
       while (isdigit(*end) || *end == '-') { end++; }
       if ((len = end - start)) {
-	buf = SEN_MALLOC(len + 1);
-	memcpy(buf, start, len);
-	buf[len] = '\0';
-	q->escalation_threshold = atoi(buf);
-	SEN_FREE(buf);
+        if (!(buf = SEN_MALLOC(len + 1))) { SEN_LOG(sen_log_alert, "get_pragma malloc fail"); return; }
+        memcpy(buf, start, len);
+        buf[len] = '\0';
+        q->escalation_threshold = atoi(buf);
+        SEN_FREE(buf);
       }
       if (*end == ',') {
-	start = ++end;
-	while (*end && *end != SEN_QUERY_PREFIX && !sen_isspace(end, q->encoding)) {
-	  end++;
-	}
-	if ((len = end - start)) {
-	  buf = SEN_MALLOC(len + 1);
-	  memcpy(buf, start, len);
-	  buf[len] = '\0';
-	  q->escalation_decaystep = atoi(buf);
-	  SEN_FREE(buf);
-	}
+        start = ++end;
+        while (*end && *end != SEN_QUERY_PREFIX && !sen_isspace(end, q->encoding)) {
+          end++;
+        }
+        if ((len = end - start)) {
+          if (!(buf = SEN_MALLOC(len + 1))) { SEN_LOG(sen_log_alert, "get_pragma malloc fail"); return; }
+          memcpy(buf, start, len);
+          buf[len] = '\0';
+          q->escalation_decaystep = atoi(buf);
+          SEN_FREE(buf);
+        }
       }
       q->cur = end;
       break;
     case 'D' :
       start = ++end;
       while (*end && *end != SEN_QUERY_PREFIX && !sen_isspace(end, q->encoding)) {
-	end++;
+        end++;
       }
       if (end > start) {
-	switch (*start) {
-	case 'O' :
-	  q->default_op = sen_sel_or;
-	  break;
-	case SEN_QUERY_AND :
-	  q->default_op = sen_sel_and;
-	  break;
-	case SEN_QUERY_BUT :
-	  q->default_op = sen_sel_but;
-	  break;
-	case SEN_QUERY_ADJ_INC :
-	  q->default_op = sen_sel_adjust;
-	  break;
-	}
+        switch (*start) {
+        case 'O' :
+          q->default_op = sen_sel_or;
+          break;
+        case SEN_QUERY_AND :
+          q->default_op = sen_sel_and;
+          break;
+        case SEN_QUERY_BUT :
+          q->default_op = sen_sel_but;
+          break;
+        case SEN_QUERY_ADJ_INC :
+          q->default_op = sen_sel_adjust;
+          break;
+        }
       }
       q->cur = end;
       break;
@@ -392,11 +391,13 @@ get_pragma(sen_query *q)
 }
 
 sen_query *
-sen_query_open(const char *str, sen_sel_operator default_op,
-	       int max_exprs, sen_encoding encoding)
+sen_query_open(const char *str, unsigned int str_len,
+               sen_sel_operator default_op, int max_exprs, sen_encoding encoding)
 {
+  // TODO: use str_len
   sen_query *q;
   if (!(q = SEN_MALLOC(sizeof(sen_query) + max_exprs * sizeof(cons)))) {
+    SEN_LOG(sen_log_alert, "sen_query_open malloc fail");
     return NULL;
   }
   q->str = str;
@@ -411,7 +412,11 @@ sen_query_open(const char *str, sen_sel_operator default_op,
   q->weight_offset = 0;
   get_pragma(q);
   q->expr = get_expr(q);
-  q->buffer = SEN_MALLOC(q->max_size + 1);
+  if (!(q->buffer = SEN_MALLOC(q->max_size + 1))) {
+    SEN_FREE(q);
+    SEN_LOG(sen_log_alert, "sen_query_open buffer malloc fail");
+    return NULL;
+  }
   return q;
 }
 
@@ -421,7 +426,7 @@ sen_query_rest(sen_query *q)
   return q->cur;
 }
 
-sen_rc 
+sen_rc
 sen_query_close(sen_query *q)
 {
   if (!q) { return sen_invalid_argument; }
@@ -445,36 +450,40 @@ exec_query(sen_index *i, sen_query *q, cons *c, sen_records *r, sen_sel_operator
     opt.vector_size = c->weight + q->weight_offset;
     opt.func = NULL;
     opt.func_arg = NULL;
-    sen_log("mode=%d option=%d w=%d o=%d", opt.mode, c->option, c->weight, q->weight_offset);
-    sen_index_select(i, str, r, op, &opt);
+    SEN_LOG(sen_log_debug, "mode=%d option=%d w=%d o=%d", opt.mode, c->option, c->weight, q->weight_offset);
+    if (sen_index_select(i, str, strlen(str), r, op, &opt)) {
+      SEN_LOG(sen_log_error, "sen_index_select on exec_query failed !");
+      return;
+    }
   } else {
     cons *token = c->u.expr.car;
     if (token) {
       sen_sel_operator t_op;
       sen_records *s = n ? sen_records_open(sen_rec_document, sen_rec_none, 0) : r;
+      if (!s) { SEN_LOG(sen_log_error, "records is null on exec_query"); return; }
       for (t_op = sen_sel_or;; t_op = token->op) {
-	exec_query(i, q, token, s, t_op);
-	if (!(token = token->cdr)) { break; }
+        exec_query(i, q, token, s, t_op);
+        if (!(token = token->cdr)) { break; }
       }
       if (n) {
-	switch (op) {
-	case sen_sel_or :
-	  if (!sen_records_union(r, s)) { sen_records_close(s); }
-	  break;
-	case sen_sel_and :
-	  if (!sen_records_intersect(r, s)) { sen_records_close(s); }
-	  break;
-	case sen_sel_but :
-	  if (!sen_records_subtract(r, s)) { sen_records_close(s); }
-	  break;
-	  /* todo: adjust
-	case sen_sel_adjust :
-	  break;
-	  */
-	default :
-	  sen_records_close(s);
-	  break;
-	}
+        switch (op) {
+        case sen_sel_or :
+          if (!sen_records_union(r, s)) { sen_records_close(s); }
+          break;
+        case sen_sel_and :
+          if (!sen_records_intersect(r, s)) { sen_records_close(s); }
+          break;
+        case sen_sel_but :
+          if (!sen_records_subtract(r, s)) { sen_records_close(s); }
+          break;
+          /* todo: adjust
+        case sen_sel_adjust :
+          break;
+          */
+        default :
+          sen_records_close(s);
+          break;
+        }
       }
     }
   }
@@ -491,19 +500,19 @@ sen_query_exec(sen_index *i, sen_query *q, sen_records *r, sen_sel_operator op)
   if (p >= 0 || (-p & 1)) {
     q->default_mode = sen_sel_exact;
     exec_query(i, q, q->expr, r, op);
-    sen_log("hits(exact)=%d", sen_records_nhits(r));
+    SEN_LOG(sen_log_info, "hits(exact)=%d", sen_records_nhits(r));
   }
   if ((p >= 0) ? (p >= sen_records_nhits(r)) : (-p & 2)) {
     q->weight_offset -= q->escalation_decaystep;
     q->default_mode = sen_sel_unsplit;
     exec_query(i, q, q->expr, r, op);
-    sen_log("hits(unsplit)=%d", sen_records_nhits(r));
+    SEN_LOG(sen_log_info, "hits(unsplit)=%d", sen_records_nhits(r));
   }
   if ((p >= 0) ? (p >= sen_records_nhits(r)) : (-p & 4)) {
     q->weight_offset -= q->escalation_decaystep;
     q->default_mode = sen_sel_partial;
     exec_query(i, q, q->expr, r, op);
-    sen_log("hits(partial)=%d", sen_records_nhits(r));
+    SEN_LOG(sen_log_info, "hits(partial)=%d", sen_records_nhits(r));
   }
   return sen_success;
 }
