@@ -4,12 +4,12 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
@@ -42,13 +42,13 @@ sen_bm_check_sjis(const unsigned char *x, const size_t y)
 {
   const unsigned char *p;
   for (p = x + y - 1; p >= x; p--)
-    if ((*p < 0x81) || (*p > 0x9F && *p < 0xE0) || (*p > 0xEF))
+    if ((*p < 0x81) || (*p > 0x9F && *p < 0xE0) || (*p > 0xFC))
       break;
   return (int) ((x + y - p) % 2);
 }
 
 void
-sen_bm_suffixes(const unsigned char *x, size_t m, size_t * suff)
+sen_bm_suffixes(const unsigned char *x, size_t m, size_t *suff)
 {
   size_t f, g;
   intptr_t i;
@@ -70,7 +70,7 @@ sen_bm_suffixes(const unsigned char *x, size_t m, size_t * suff)
 }
 
 void
-sen_bm_preBmBc(const unsigned char *x, size_t m, size_t * bmBc)
+sen_bm_preBmBc(const unsigned char *x, size_t m, size_t *bmBc)
 {
   size_t i;
   for (i = 0; i < ASIZE; ++i)
@@ -122,7 +122,7 @@ sen_bm_preBmBc(const unsigned char *x, size_t m, size_t * bmBc)
 }
 
 inline void
-sen_bm_TunedBM(_snip_cond * cond, sen_nstr * object, size_t n, sen_encoding encoding)
+sen_bm_TunedBM(_snip_cond *cond, sen_nstr *object, size_t n, sen_encoding encoding)
 {
   register unsigned char *limit, ck;
   register const unsigned char *p, *cp;
@@ -235,7 +235,7 @@ count_mapped_chars(const char *str, const char *end)
 }
 
 sen_rc
-snip_cond_close(_snip_cond * cond)
+snip_cond_close(_snip_cond *cond)
 {
   if (!cond) {
     return sen_invalid_argument;
@@ -247,8 +247,10 @@ snip_cond_close(_snip_cond * cond)
 }
 
 sen_rc
-sen_snip_add_cond(sen_snip * snip, const char *keyword,
-                  const char *opentag, const char *closetag)
+sen_snip_add_cond(sen_snip *snip,
+                  const char *keyword, unsigned int keyword_len,
+                  const char *opentag, unsigned int opentag_len,
+                  const char *closetag, unsigned int closetag_len)
 {
   _snip_cond *cond = snip->cond + snip->cond_len;
 
@@ -258,14 +260,14 @@ sen_snip_add_cond(sen_snip * snip, const char *keyword,
   memset(cond, 0, sizeof(_snip_cond));
   if (opentag) {
     cond->opentag = opentag;
-    cond->opentag_len = strlen(opentag);
+    cond->opentag_len = opentag_len;
   } else {
     cond->opentag = snip->defaultopentag;
     cond->opentag_len = snip->defaultopentag_len;
   }
   if (closetag) {
     cond->closetag = closetag;
-    cond->closetag_len = strlen(closetag);
+    cond->closetag_len = closetag_len;
   } else {
     cond->closetag = snip->defaultclosetag;
     cond->closetag_len = snip->defaultclosetag_len;
@@ -273,17 +275,22 @@ sen_snip_add_cond(sen_snip * snip, const char *keyword,
 
   if (snip->flags & SEN_SNIP_NORMALIZE) {
     cond->keyword = sen_nstr_open(keyword, snip->encoding, SEN_STR_REMOVEBLANK);        /* normalize dup */
+    if (!cond->keyword) {
+      SEN_LOG(sen_log_alert, "sen_nstr_open on sen_snip_add_cond failed !");
+      return sen_memory_exhausted;
+    }
   } else {
     sen_nstr *nstr;
     if (!(nstr = SEN_MALLOC(sizeof(sen_nstr)))) {
-      goto exit;
+      SEN_LOG(sen_log_alert, "sen_nstr_open on sen_snip_add_cond failed !");
+      return sen_memory_exhausted;
     }
     nstr->norm = SEN_STRDUP(keyword);   /* dup, the label is 'norm' but not normalized  */
     nstr->ctypes = NULL;
     nstr->checks = NULL;
     cond->keyword = nstr;
   }
-  cond->keyword_len = strlen(cond->keyword->norm);
+  cond->keyword_len = strlen(cond->keyword->norm); /* byte length, not cond->keyword->length */
   if (!cond->keyword_len || cond->keyword_len > snip->width) {
     snip_cond_close(cond);
     return sen_invalid_argument;
@@ -295,10 +302,6 @@ sen_snip_add_cond(sen_snip * snip, const char *keyword,
 
   snip->cond_len++;
   return sen_success;
-
-exit:
-  snip_cond_close(cond);
-  return sen_memory_exhausted;
 }
 
 size_t
@@ -325,9 +328,11 @@ sen_snip_find_firstbyte(const char *string, sen_encoding encoding, size_t offset
 }
 
 sen_snip *
-sen_snip_open(sen_encoding encoding, int flags, size_t width,
-              unsigned int max_results, const char *defaultopentag,
-              const char *defaultclosetag, sen_snip_mapping * mapping)
+sen_snip_open(sen_encoding encoding, int flags, unsigned int width,
+              unsigned int max_results,
+              const char *defaultopentag, unsigned int defaultopentag_len,
+              const char *defaultclosetag, unsigned int defaultclosetag_len,
+              sen_snip_mapping *mapping)
 {
   sen_snip *ret = NULL;
   if (!(ret = SEN_MALLOC(sizeof(sen_snip))) || max_results > MAX_SNIP_RESULT_COUNT
@@ -340,8 +345,8 @@ sen_snip_open(sen_encoding encoding, int flags, size_t width,
   ret->max_results = max_results;
   ret->defaultopentag = defaultopentag;
   ret->defaultclosetag = defaultclosetag;
-  ret->defaultopentag_len = strlen(defaultopentag);
-  ret->defaultclosetag_len = strlen(defaultclosetag);
+  ret->defaultopentag_len = defaultopentag_len;
+  ret->defaultclosetag_len = defaultclosetag_len;
   ret->cond_len = 0;
   ret->mapping = mapping;
   ret->nstr = NULL;
@@ -352,7 +357,7 @@ sen_snip_open(sen_encoding encoding, int flags, size_t width,
 }
 
 sen_rc
-exec_clean(sen_snip * snip)
+exec_clean(sen_snip *snip)
 {
   _snip_cond *cond;
   if (snip->nstr) {
@@ -374,7 +379,7 @@ exec_clean(sen_snip * snip)
 }
 
 sen_rc
-sen_snip_close(sen_snip * snip)
+sen_snip_close(sen_snip *snip)
 {
   size_t i;
   if (!snip) {
@@ -403,7 +408,7 @@ sen_fakenstr_open(const char *string, size_t string_blen, sen_encoding encoding,
   nstr->norm = SEN_STRDUP(string);
   nstr->ctypes = NULL;
   if (!(nstr->checks = (int_least8_t *) SEN_MALLOC(sizeof(int_least8_t) * string_blen))) {
-    free(nstr);
+    SEN_FREE(nstr);
     return NULL;
   }
 
@@ -458,11 +463,10 @@ sen_fakenstr_open(const char *string, size_t string_blen, sen_encoding encoding,
 }
 
 sen_rc
-sen_snip_exec(sen_snip * snip, const char *string, unsigned int *nresults,
-              size_t * max_tagged_len)
+sen_snip_exec(sen_snip *snip, const char *string, unsigned int string_len,
+              unsigned int *nresults, unsigned int *max_tagged_len)
 {
-  sen_rc ret = sen_success;
-  size_t nstr_blen, string_blen, i;
+  size_t nstr_blen, i;
 
   exec_clean(snip);
 
@@ -470,17 +474,18 @@ sen_snip_exec(sen_snip * snip, const char *string, unsigned int *nresults,
   if (!string || !*string) {
     return sen_invalid_argument;
   }
-  string_blen = strlen(string);
   if (snip->flags & SEN_SNIP_NORMALIZE) {
     snip->nstr =
       sen_nstr_open(string, snip->encoding, SEN_STR_WITH_CHECKS | SEN_STR_REMOVEBLANK);
   } else {
     snip->nstr =
-      sen_fakenstr_open(string, string_blen, snip->encoding,
+      sen_fakenstr_open(string, string_len, snip->encoding,
                         SEN_STR_WITH_CHECKS | SEN_STR_REMOVEBLANK);
   }
   if (!snip->nstr) {
-    goto exit;
+    exec_clean(snip);
+    SEN_LOG(sen_log_alert, "sen_nstr_open on sen_snip_exec failed !");
+    return sen_memory_exhausted;
   }
   nstr_blen = strlen(snip->nstr->norm); /* byte length, not nstr->length */
 
@@ -583,12 +588,12 @@ sen_snip_exec(sen_snip * snip, const char *string, unsigned int *nresults,
         snip_result->start_offset =
           MAX(MIN
               ((snip_result->start_offset + last_end_offset - snip->width) / 2,
-               string_blen - snip->width), last_last_end_offset);
+               string_len - snip->width), last_last_end_offset);
       }
       snip_result->start_offset =
         sen_snip_find_firstbyte(string, snip->encoding, snip_result->start_offset, 1);
 
-      snip_result->end_offset = MIN(string_blen, snip_result->start_offset + snip->width);
+      snip_result->end_offset = MIN(string_len, snip_result->start_offset + snip->width);
 
       snip_result->end_offset =
         sen_snip_find_firstbyte(string, snip->encoding, snip_result->end_offset, -1);
@@ -624,14 +629,11 @@ sen_snip_exec(sen_snip * snip, const char *string, unsigned int *nresults,
 
   snip->max_tagged_len = *max_tagged_len;
 
-  return ret;
-exit:
-  exec_clean(snip);
-  return ret;
+  return sen_success;
 }
 
 sen_rc
-sen_snip_get_result(sen_snip * snip, const unsigned int index, char *result)
+sen_snip_get_result(sen_snip *snip, const unsigned int index, char *result, unsigned int *result_len)
 {
   char *p;
   size_t i, j, k;
@@ -705,7 +707,8 @@ sen_snip_get_result(sen_snip * snip, const unsigned int index, char *result)
   }
   *p = '\0';
 
-  assert((size_t) (p - result) <= snip->max_tagged_len);
+  if(result_len) { *result_len = (unsigned int)(p - result); }
+  assert((unsigned int)(p - result) <= snip->max_tagged_len);
 
   return sen_success;
 }
