@@ -36,6 +36,7 @@ prefix_init(void)
   FILE *fp;
   char buffer[4];
   prefix = sen_set_open(2, sizeof(int), 0);
+  if (!prefix) { SEN_LOG(sen_log_alert, "sen_set_open on prefix_init failed !"); return; }
   if ((fp = fopen(PREFIX_PATH, "r"))) {
     for (i = 0; i < N_PREFIX; i++) {
       if (!fgets(buffer, 4, fp)) { break; }
@@ -53,6 +54,7 @@ suffix_init(void)
   FILE *fp;
   char buffer[4];
   suffix = sen_set_open(2, 0, 0);
+  if (!suffix) { SEN_LOG(sen_log_alert, "sen_set_open on suffix_init failed !"); return; }
   if ((fp = fopen(SUFFIX_PATH, "r"))) {
     for (i = N_SUFFIX; i; i--) {
       if (!fgets(buffer, 4, fp)) { break; }
@@ -88,7 +90,7 @@ sen_str_get_prefix_order(const char *str)
 /*
 static void
 default_logger_func(int level, const char *time,
-		    const char *title, const char *msg, const char *location)
+                    const char *title, const char *msg, const char *location)
 {
   fprintf(stderr, "%s|%s %s %s\n", time, location, msg, note);
 }
@@ -96,22 +98,23 @@ default_logger_func(int level, const char *time,
 
 static void
 default_logger_func(int level, const char *time, const char *title,
-		    const char *msg, const char *location, void *func_arg)
+                    const char *msg, const char *location, void *func_arg)
 {
   static FILE *fp = NULL;
+  const char slev[] = " EACewnid-";
   if (!fp) {
     // mutex_lock
     fp = fopen(SENNA_HOME PATH_SEPARATOR "log" PATH_SEPARATOR "senna.log", "a");
     // mutex_unlock
   }
   if (fp) {
-    fprintf(fp, "%s|%s %s %s\n", time, title, msg, location);
+    fprintf(fp, "%s|%c|%s %s %s\n", time, *(slev + level), title, msg, location);
     fflush(fp);
   }
 }
 
 static sen_logger_info default_logger = {
-  sen_log_info,
+  SEN_LOG_DEFAULT_LEVEL,
   SEN_LOG_TIME|SEN_LOG_MESSAGE,
   default_logger_func
 };
@@ -141,7 +144,7 @@ sen_logger_pass(sen_log_level level)
 
 void
 sen_logger_put(sen_log_level level, 
-	       const char *file, int line, const char *func, char *fmt, ...)
+               const char *file, int line, const char *func, char *fmt, ...)
 {
   if (level <= sen_logger->max_level) {
     char tbuf[TBUFSIZE];
@@ -156,8 +159,8 @@ sen_logger_put(sen_log_level level,
       ltm = localtime(&t);
       _ftime(&tb);
       snprintf(tbuf, TBUFSIZE - 1, "%02d/%02d:%02d:%02d:%02d.%03d",
-	       ltm->tm_mon + 1, ltm->tm_mday,
-	       ltm->tm_hour, ltm->tm_min, ltm->tm_sec, tb.millitm);
+               ltm->tm_mon + 1, ltm->tm_mday,
+               ltm->tm_hour, ltm->tm_min, ltm->tm_sec, tb.millitm);
       tbuf[TBUFSIZE - 1] = '\0';
 #else /* WIN32 */
       struct timeval tv;
@@ -165,8 +168,8 @@ sen_logger_put(sen_log_level level,
       gettimeofday(&tv, NULL);
       localtime_r((time_t *)&tv.tv_sec, &tm);
       snprintf(tbuf, TBUFSIZE - 1, "%02d/%02d:%02d:%02d:%02d.%06d",
-	       tm.tm_mon + 1, tm.tm_mday,
-	       tm.tm_hour, tm.tm_min, tm.tm_sec, (int) tv.tv_usec);
+               tm.tm_mon + 1, tm.tm_mday,
+               tm.tm_hour, tm.tm_min, tm.tm_sec, (int) tv.tv_usec);
       tbuf[TBUFSIZE - 1] = '\0';
 #endif /* WIN32 */
     } else {
@@ -183,7 +186,7 @@ sen_logger_put(sen_log_level level,
     }
     if (sen_logger->flags & SEN_LOG_LOCATION) {
       snprintf(lbuf, LBUFSIZE - 1, "%04x %s:%d %s()",
-	       getpid(), file, line, func);
+               getpid(), file, line, func);
       lbuf[LBUFSIZE - 1] = '\0';
     } else {
       lbuf[0] = '\0';
@@ -205,7 +208,7 @@ sen_malloc(size_t size, const char* file, int line)
   void *res;
   res = malloc(size);
   if (res) { alloc_count++; }
-  if (!res) { sen_log("malloc(%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count); }
+  if (!res) { SEN_LOG(sen_log_alert, "malloc fail (%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count); }
   return res;
 }
 
@@ -215,7 +218,7 @@ sen_calloc(size_t size, const char* file, int line)
   void *res;
   res = calloc(size, 1);
   if (res) { alloc_count++; }
-  if (!res) { sen_log("calloc(%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count); }
+  if (!res) { SEN_LOG(sen_log_alert, "calloc fail (%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count); }
   return res;
 }
 
@@ -224,7 +227,7 @@ sen_free(void *ptr, const char* file, int line)
 {
   free(ptr);
   if (ptr) { alloc_count--; }
-  if (!ptr) { sen_log("free(%p) (%s:%d) <%d>", ptr, file, line, alloc_count); }
+  if (!ptr) { SEN_LOG(sen_log_alert, "free fail (%p) (%s:%d) <%d>", ptr, file, line, alloc_count); }
 }
 
 void *
@@ -233,7 +236,7 @@ sen_realloc(void *ptr, size_t size, const char* file, int line)
   void *res;
   res = realloc(ptr, size);
   if (res) { alloc_count++; }
-  sen_log("realloc(%p,%d)=%p (%s:%d) <%d>", ptr, size, res, file, line, alloc_count);
+  if (!res) { SEN_LOG(sen_log_alert, "realloc fail (%p,%d)=%p (%s:%d) <%d>", ptr, size, res, file, line, alloc_count); }
   return res;
 }
 
@@ -243,10 +246,18 @@ sen_strdup(const char *s, const char* file, int line)
   char *res;
   res = strdup(s);
   if (res) { alloc_count++; }
-  sen_log("strdup(%p)=%p (%s:%d) <%d>", s, res, file, line, alloc_count);
+  if (!res) { SEN_LOG(sen_log_alert, "strdup(%p)=%p (%s:%d) <%d>", s, res, file, line, alloc_count); }
   return res;
 }
 
+void
+sen_assert(int cond, const char* file, int line, const char* func)
+{
+  if (!cond) {
+    SEN_LOG(sen_log_error, "ASSERT fail on %s %s:%d", func, file, line);
+  }
+}
+
 static unsigned char symbol[] = {
   ',', '.', 0, ':', ';', '?', '!', 0, 0, 0, '`', 0, '^', '~', '_', 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, '-', '-', '/', '\\', 0, 0, '|', 0, 0, 0, '\'', 0,
@@ -289,12 +300,15 @@ normalize_euc(sen_nstr *nstr)
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
     if (!(nstr->checks = SEN_MALLOC(size * 2 + 1))) {
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
   ch = nstr->checks;
   if (nstr->flags & SEN_STR_WITH_CTYPES) {
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
+      SEN_FREE(nstr->checks);
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
@@ -302,173 +316,173 @@ normalize_euc(sen_nstr *nstr)
   for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; *s; s++) {
     if ((*s & 0x80)) {
       if ((*(s + 1) & 0x80)) {
-	unsigned char c1 = *s++, c2 = *s, c3 = 0;
-	switch (c1 >> 4) {
-	case 0x08 :
-	  if (c1 == 0x8e && 0xa0 <= c2 && c2 <= 0xdf) {
-	    uint16_t c = hankana[c2 - 0xa0];
-	    switch (c) {
-	    case 0xa1ab :
-	      if (d > d0 + 1 && d[-2] == 0xa5
-		  && 0xa6 <= d[-1] && d[-1] <= 0xdb && (b = dakuten[d[-1] - 0xa6])) {
-		*(d - 1) = b;
-		if (ch) { ch[-1] += 2; s_ += 2; }
-		continue;
-	      } else {
-		*d++ = c >> 8; *d = c & 0xff;
-	      }
-	      break;
-	    case 0xa1eb :
-	      if (d > d0 + 1 && d[-2] == 0xa5
-		  && 0xcf <= d[-1] && d[-1] <= 0xdb && (b = handaku[d[-1] - 0xcf])) {
-		*(d - 1) = b;
-		if (ch) { ch[-1] += 2; s_ += 2; }
-		continue;
-	      } else {
-		*d++ = c >> 8; *d = c & 0xff;
-	      }
-	      break;
-	    default :
-	      *d++ = c >> 8; *d = c & 0xff;
-	      break;
-	    }
-	    ctype = sen_str_katakana;
-	  } else {
-	    *d++ = c1; *d = c2;
-	    ctype = sen_str_others;
-	  }
-	  break;
-	case 0x09 :
-	  *d++ = c1; *d = c2;
-	  ctype = sen_str_others;
-	  break;
-	case 0x0a :
-	  switch (c1 & 0x0f) {
-	  case 1 :
-	    switch (c2) {
-	    case 0xbc :
-	      *d++ = c1; *d = c2;
-	      ctype = sen_str_katakana;
-	      break;
-	    case 0xb9 :
-	      *d++ = c1; *d = c2;
-	      ctype = sen_str_kanji;
-	      break;
-	    case 0xa1 :
-	      if (removeblankp) {
-		if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-		continue;
-	      } else {
-		*d = ' ';
-		ctype = SEN_NSTR_BLANK|sen_str_symbol;
-	      }
-	      break;
-	    default :
-	      if (c2 >= 0xa4 && (c3 = symbol[c2 - 0xa4])) {
-		*d = c3;
-		ctype = sen_str_symbol;
-	      } else {
-		*d++ = c1; *d = c2;
-		ctype = sen_str_others;
-	      }
-	      break;
-	    }
-	    break;
-	  case 2 :
-	    *d++ = c1; *d = c2;
-	    ctype = sen_str_symbol;
-	    break;
-	  case 3 :
-	    c3 = c2 - 0x80;
-	    if ('a' <= c3 && c3 <= 'z') {
-	      ctype = sen_str_alpha;
-	      *d = c3;
-	    } else if ('A' <= c3 && c3 <= 'Z') {
-	      ctype = sen_str_alpha;
-	      *d = c3 + 0x20;
-	    } else if ('0' <= c3 && c3 <= '9') {
-	      ctype = sen_str_digit;
-	      *d = c3;
-	    } else {
-	      ctype = sen_str_others;
-	      *d++ = c1; *d = c2;
-	    }
-	    break;
-	  case 4 :
-	    *d++ = c1; *d = c2;
-	    ctype = sen_str_hiragana;
-	    break;
-	  case 5 :
-	    *d++ = c1; *d = c2;
-	    ctype = sen_str_katakana;
-	    break;
-	  case 6 :
-	  case 7 :
-	  case 8 :
-	    *d++ = c1; *d = c2;
-	    ctype = sen_str_symbol;
-	    break;
-	  default :
-	    *d++ = c1; *d = c2;
-	    ctype = sen_str_others;
-	    break;
-	  }
-	  break;
-	default :
-	  *d++ = c1; *d = c2;
-	  ctype = sen_str_kanji;
-	  break;
-	}
+        unsigned char c1 = *s++, c2 = *s, c3 = 0;
+        switch (c1 >> 4) {
+        case 0x08 :
+          if (c1 == 0x8e && 0xa0 <= c2 && c2 <= 0xdf) {
+            uint16_t c = hankana[c2 - 0xa0];
+            switch (c) {
+            case 0xa1ab :
+              if (d > d0 + 1 && d[-2] == 0xa5
+                  && 0xa6 <= d[-1] && d[-1] <= 0xdb && (b = dakuten[d[-1] - 0xa6])) {
+                *(d - 1) = b;
+                if (ch) { ch[-1] += 2; s_ += 2; }
+                continue;
+              } else {
+                *d++ = c >> 8; *d = c & 0xff;
+              }
+              break;
+            case 0xa1eb :
+              if (d > d0 + 1 && d[-2] == 0xa5
+                  && 0xcf <= d[-1] && d[-1] <= 0xdb && (b = handaku[d[-1] - 0xcf])) {
+                *(d - 1) = b;
+                if (ch) { ch[-1] += 2; s_ += 2; }
+                continue;
+              } else {
+                *d++ = c >> 8; *d = c & 0xff;
+              }
+              break;
+            default :
+              *d++ = c >> 8; *d = c & 0xff;
+              break;
+            }
+            ctype = sen_str_katakana;
+          } else {
+            *d++ = c1; *d = c2;
+            ctype = sen_str_others;
+          }
+          break;
+        case 0x09 :
+          *d++ = c1; *d = c2;
+          ctype = sen_str_others;
+          break;
+        case 0x0a :
+          switch (c1 & 0x0f) {
+          case 1 :
+            switch (c2) {
+            case 0xbc :
+              *d++ = c1; *d = c2;
+              ctype = sen_str_katakana;
+              break;
+            case 0xb9 :
+              *d++ = c1; *d = c2;
+              ctype = sen_str_kanji;
+              break;
+            case 0xa1 :
+              if (removeblankp) {
+                if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+                continue;
+              } else {
+                *d = ' ';
+                ctype = SEN_NSTR_BLANK|sen_str_symbol;
+              }
+              break;
+            default :
+              if (c2 >= 0xa4 && (c3 = symbol[c2 - 0xa4])) {
+                *d = c3;
+                ctype = sen_str_symbol;
+              } else {
+                *d++ = c1; *d = c2;
+                ctype = sen_str_others;
+              }
+              break;
+            }
+            break;
+          case 2 :
+            *d++ = c1; *d = c2;
+            ctype = sen_str_symbol;
+            break;
+          case 3 :
+            c3 = c2 - 0x80;
+            if ('a' <= c3 && c3 <= 'z') {
+              ctype = sen_str_alpha;
+              *d = c3;
+            } else if ('A' <= c3 && c3 <= 'Z') {
+              ctype = sen_str_alpha;
+              *d = c3 + 0x20;
+            } else if ('0' <= c3 && c3 <= '9') {
+              ctype = sen_str_digit;
+              *d = c3;
+            } else {
+              ctype = sen_str_others;
+              *d++ = c1; *d = c2;
+            }
+            break;
+          case 4 :
+            *d++ = c1; *d = c2;
+            ctype = sen_str_hiragana;
+            break;
+          case 5 :
+            *d++ = c1; *d = c2;
+            ctype = sen_str_katakana;
+            break;
+          case 6 :
+          case 7 :
+          case 8 :
+            *d++ = c1; *d = c2;
+            ctype = sen_str_symbol;
+            break;
+          default :
+            *d++ = c1; *d = c2;
+            ctype = sen_str_others;
+            break;
+          }
+          break;
+        default :
+          *d++ = c1; *d = c2;
+          ctype = sen_str_kanji;
+          break;
+        }
       } else {
-	/* skip invalid character */
-	continue;
+        /* skip invalid character */
+        continue;
       }
     } else {
       unsigned char c = *s;
       switch (c >> 4) {
       case 0 :
       case 1 :
-	/* skip unprintable ascii */
-	if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-	continue;
+        /* skip unprintable ascii */
+        if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+        continue;
       case 2 :
-	if (c == 0x20) {
-	  if (removeblankp) {
-	    if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-	    continue;
-	  } else {
-	    *d = ' ';
-	    ctype = SEN_NSTR_BLANK|sen_str_symbol;
-	  }
-	} else {
-	  *d = c;
-	  ctype = sen_str_symbol;
-	}
-	break;
+        if (c == 0x20) {
+          if (removeblankp) {
+            if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+            continue;
+          } else {
+            *d = ' ';
+            ctype = SEN_NSTR_BLANK|sen_str_symbol;
+          }
+        } else {
+          *d = c;
+          ctype = sen_str_symbol;
+        }
+        break;
       case 3 :
-	*d = c;
-	ctype = (c <= 0x39) ? sen_str_digit : sen_str_symbol;
-	break;
+        *d = c;
+        ctype = (c <= 0x39) ? sen_str_digit : sen_str_symbol;
+        break;
       case 4 :
-	*d = ('A' <= c) ? c + 0x20 : c;
-	ctype = (c == 0x40) ? sen_str_symbol : sen_str_alpha;
-	break;
+        *d = ('A' <= c) ? c + 0x20 : c;
+        ctype = (c == 0x40) ? sen_str_symbol : sen_str_alpha;
+        break;
       case 5 :
-	*d = (c <= 'Z') ? c + 0x20 : c;
-	ctype = (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
-	break;
+        *d = (c <= 'Z') ? c + 0x20 : c;
+        ctype = (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
+        break;
       case 6 :
-	*d = c;
-	ctype = (c == 0x60) ? sen_str_symbol : sen_str_alpha;
-	break;
+        *d = c;
+        ctype = (c == 0x60) ? sen_str_symbol : sen_str_alpha;
+        break;
       case 7 :
-	*d = c;
-	ctype = (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
-	break;
+        *d = c;
+        ctype = (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
+        break;
       default :
-	*d = c;
-	ctype = sen_str_others;
-	break;
+        *d = c;
+        ctype = sen_str_others;
+        break;
       }
     }
     d++;
@@ -543,12 +557,15 @@ normalize_utf8(sen_nstr *nstr)
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
     if (!(nstr->checks = SEN_MALLOC(size * 5 + 1))) { /* todo: realloc unless enough */
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
   ch = nstr->checks;
   if (nstr->flags & SEN_STR_WITH_CTYPES) {
     if (!(nstr->ctypes = SEN_MALLOC(size * 3 + 1))) { /* todo: realloc unless enough */
+      SEN_FREE(nstr->checks);
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
@@ -566,8 +583,8 @@ normalize_utf8(sen_nstr *nstr)
       pe = p + strlen((char *)p);
       if (cp) { cp--; }
       if (ch) {
-	ch -= (d - d_);
-	s_ -= ls_;
+        ch -= (d - d_);
+        s_ -= ls_;
       }
       d = d_;
       length--;
@@ -575,24 +592,24 @@ normalize_utf8(sen_nstr *nstr)
     for (; *p && p < pe; p += lp) {
       lp = sen_str_charlen((char *)p, sen_enc_utf8);
       if ((*p == ' ' && removeblankp)
-	  || *p < 0x20 	/* skip unprintable ascii */ ) {
-	if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+          || *p < 0x20  /* skip unprintable ascii */ ) {
+        if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
       } else {
-	memcpy(d, p, lp);
-	d_ = d;
-	d += lp;
-	length++;
-	if (cp) { *cp++ = getctype_utf8(p); }
-	if (ch) {
-	  size_t i;
-	  if (s_ == s + ls) {
-	    *ch++ = -1;
-	  } else {
-	    *ch++ = (int_least8_t)(s + ls - s_);
-	    s_ = s + ls;
-	  }
-	  for (i = lp; i > 1; i--) { *ch++ = 0; }
-	}
+        memcpy(d, p, lp);
+        d_ = d;
+        d += lp;
+        length++;
+        if (cp) { *cp++ = getctype_utf8(p); }
+        if (ch) {
+          size_t i;
+          if (s_ == s + ls) {
+            *ch++ = -1;
+          } else {
+            *ch++ = (int_least8_t)(s + ls - s_);
+            s_ = s + ls;
+          }
+          for (i = lp; i > 1; i--) { *ch++ = 0; }
+        }
       }
     }
     ls_ = ls;
@@ -638,12 +655,15 @@ normalize_sjis(sen_nstr *nstr)
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
     if (!(nstr->checks = SEN_MALLOC(size * 2 + 1))) {
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
   ch = nstr->checks;
   if (nstr->flags & SEN_STR_WITH_CTYPES) {
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
+      SEN_FREE(nstr->checks);
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
@@ -651,164 +671,164 @@ normalize_sjis(sen_nstr *nstr)
   for (s = s_ = (unsigned char *) nstr->orig, d = d_ = d0; *s; s++) {
     if ((*s & 0x80)) {
       if (0xa0 <= *s && *s <= 0xdf) {
-	uint16_t c = hankana[*s - 0xa0];
-	switch (c) {
-	case 0x814a :
-	  if (d > d0 + 1 && d[-2] == 0x83
-	      && 0x45 <= d[-1] && d[-1] <= 0x7a && (b = dakuten[d[-1] - 0x45])) {
-	    *(d - 1) = b;
-	    if (ch) { ch[-1]++; s_++; }
-	    continue;
-	  } else {
-	    *d++ = c >> 8; *d = c & 0xff;
-	  }
-	  break;
-	case 0x814b :
-	  if (d > d0 + 1 && d[-2] == 0x83
-	      && 0x6e <= d[-1] && d[-1] <= 0x7a && (b = handaku[d[-1] - 0x6e])) {
-	    *(d - 1) = b;
-	    if (ch) { ch[-1]++; s_++; }
-	    continue;
-	  } else {
-	    *d++ = c >> 8; *d = c & 0xff;
-	  }
-	  break;
-	default :
-	  *d++ = c >> 8; *d = c & 0xff;
-	  break;
-	}
-	ctype = sen_str_katakana;
+        uint16_t c = hankana[*s - 0xa0];
+        switch (c) {
+        case 0x814a :
+          if (d > d0 + 1 && d[-2] == 0x83
+              && 0x45 <= d[-1] && d[-1] <= 0x7a && (b = dakuten[d[-1] - 0x45])) {
+            *(d - 1) = b;
+            if (ch) { ch[-1]++; s_++; }
+            continue;
+          } else {
+            *d++ = c >> 8; *d = c & 0xff;
+          }
+          break;
+        case 0x814b :
+          if (d > d0 + 1 && d[-2] == 0x83
+              && 0x6e <= d[-1] && d[-1] <= 0x7a && (b = handaku[d[-1] - 0x6e])) {
+            *(d - 1) = b;
+            if (ch) { ch[-1]++; s_++; }
+            continue;
+          } else {
+            *d++ = c >> 8; *d = c & 0xff;
+          }
+          break;
+        default :
+          *d++ = c >> 8; *d = c & 0xff;
+          break;
+        }
+        ctype = sen_str_katakana;
       } else {
-	if (0x40 <= *(s + 1) && *(s + 1) <= 0xfc) {
-	  unsigned char c1 = *s++, c2 = *s, c3 = 0;
-	  if (0x81 <= c1 && c1 <= 0x87) {
-	    switch (c1 & 0x0f) {
-	    case 1 :
-	      switch (c2) {
-	      case 0x5b :
-		*d++ = c1; *d = c2;
-		ctype = sen_str_katakana;
-		break;
-	      case 0x58 :
-		*d++ = c1; *d = c2;
-		ctype = sen_str_kanji;
-		break;
-	      case 0x40 :
-		if (removeblankp) {
-		  if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-		  continue;
-		} else {
-		  *d = ' ';
-		  ctype = SEN_NSTR_BLANK|sen_str_symbol;
-		}
-		break;
-	      default :
-		if (0x43 <= c2 && c2 <= 0x7e && (c3 = symbol[c2 - 0x43])) {
-		  *d = c3;
-		  ctype = sen_str_symbol;
-		} else if (0x7f <= c2 && c2 <= 0x97 && (c3 = symbol[c2 - 0x44])) {
-		  *d = c3;
-		  ctype = sen_str_symbol;
-		} else {
-		  *d++ = c1; *d = c2;
-		  ctype = sen_str_others;
-		}
-		break;
-	      }
-	      break;
-	    case 2 :
-	      c3 = c2 - 0x1f;
-	      if (0x4f <= c2 && c2 <= 0x58) {
-		ctype = sen_str_digit;
-		*d = c2 - 0x1f;
-	      } else if (0x60 <= c2 && c2 <= 0x79) {
-		ctype = sen_str_alpha;
-		*d = c2 + 0x01;
-	      } else if (0x81 <= c2 && c2 <= 0x9a) {
-		ctype = sen_str_alpha;
-		*d = c2 - 0x20;
-	      } else if (0x9f <= c2 && c2 <= 0xf1) {
-		*d++ = c1; *d = c2;
-		ctype = sen_str_hiragana;
-	      } else {
-		*d++ = c1; *d = c2;
-		ctype = sen_str_others;
-	      }
-	      break;
-	    case 3 :
-	      if (0x40 <= c2 && c2 <= 0x96) {
-		*d++ = c1; *d = c2;
-		ctype = sen_str_katakana;
-	      } else {
-		*d++ = c1; *d = c2;
-		ctype = sen_str_symbol;
-	      }
-	      break;
-	    case 4 :
-	    case 7 :
-	      *d++ = c1; *d = c2;
-	      ctype = sen_str_symbol;
-	      break;
-	    default :
-	      *d++ = c1; *d = c2;
-	      ctype = sen_str_others;
-	      break;
-	    }
-	  } else {
-	    *d++ = c1; *d = c2;
-	    ctype = sen_str_kanji;
-	  }
-	} else {
-	  /* skip invalid character */
-	  continue;
-	}
+        if (0x40 <= *(s + 1) && *(s + 1) <= 0xfc) {
+          unsigned char c1 = *s++, c2 = *s, c3 = 0;
+          if (0x81 <= c1 && c1 <= 0x87) {
+            switch (c1 & 0x0f) {
+            case 1 :
+              switch (c2) {
+              case 0x5b :
+                *d++ = c1; *d = c2;
+                ctype = sen_str_katakana;
+                break;
+              case 0x58 :
+                *d++ = c1; *d = c2;
+                ctype = sen_str_kanji;
+                break;
+              case 0x40 :
+                if (removeblankp) {
+                  if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+                  continue;
+                } else {
+                  *d = ' ';
+                  ctype = SEN_NSTR_BLANK|sen_str_symbol;
+                }
+                break;
+              default :
+                if (0x43 <= c2 && c2 <= 0x7e && (c3 = symbol[c2 - 0x43])) {
+                  *d = c3;
+                  ctype = sen_str_symbol;
+                } else if (0x7f <= c2 && c2 <= 0x97 && (c3 = symbol[c2 - 0x44])) {
+                  *d = c3;
+                  ctype = sen_str_symbol;
+                } else {
+                  *d++ = c1; *d = c2;
+                  ctype = sen_str_others;
+                }
+                break;
+              }
+              break;
+            case 2 :
+              c3 = c2 - 0x1f;
+              if (0x4f <= c2 && c2 <= 0x58) {
+                ctype = sen_str_digit;
+                *d = c2 - 0x1f;
+              } else if (0x60 <= c2 && c2 <= 0x79) {
+                ctype = sen_str_alpha;
+                *d = c2 + 0x01;
+              } else if (0x81 <= c2 && c2 <= 0x9a) {
+                ctype = sen_str_alpha;
+                *d = c2 - 0x20;
+              } else if (0x9f <= c2 && c2 <= 0xf1) {
+                *d++ = c1; *d = c2;
+                ctype = sen_str_hiragana;
+              } else {
+                *d++ = c1; *d = c2;
+                ctype = sen_str_others;
+              }
+              break;
+            case 3 :
+              if (0x40 <= c2 && c2 <= 0x96) {
+                *d++ = c1; *d = c2;
+                ctype = sen_str_katakana;
+              } else {
+                *d++ = c1; *d = c2;
+                ctype = sen_str_symbol;
+              }
+              break;
+            case 4 :
+            case 7 :
+              *d++ = c1; *d = c2;
+              ctype = sen_str_symbol;
+              break;
+            default :
+              *d++ = c1; *d = c2;
+              ctype = sen_str_others;
+              break;
+            }
+          } else {
+            *d++ = c1; *d = c2;
+            ctype = sen_str_kanji;
+          }
+        } else {
+          /* skip invalid character */
+          continue;
+        }
       }
     } else {
       unsigned char c = *s;
       switch (c >> 4) {
       case 0 :
       case 1 :
-	/* skip unprintable ascii */
-	if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-	continue;
+        /* skip unprintable ascii */
+        if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+        continue;
       case 2 :
-	if (c == 0x20) {
-	  if (removeblankp) {
-	    if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-	    continue;
-	  } else {
-	    *d = ' ';
-	    ctype = SEN_NSTR_BLANK|sen_str_symbol;
-	  }
-	} else {
-	  *d = c;
-	  ctype = sen_str_symbol;
-	}
-	break;
+        if (c == 0x20) {
+          if (removeblankp) {
+            if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+            continue;
+          } else {
+            *d = ' ';
+            ctype = SEN_NSTR_BLANK|sen_str_symbol;
+          }
+        } else {
+          *d = c;
+          ctype = sen_str_symbol;
+        }
+        break;
       case 3 :
-	*d = c;
-	ctype = (c <= 0x39) ? sen_str_digit : sen_str_symbol;
-	break;
+        *d = c;
+        ctype = (c <= 0x39) ? sen_str_digit : sen_str_symbol;
+        break;
       case 4 :
-	*d = ('A' <= c) ? c + 0x20 : c;
-	ctype = (c == 0x40) ? sen_str_symbol : sen_str_alpha;
-	break;
+        *d = ('A' <= c) ? c + 0x20 : c;
+        ctype = (c == 0x40) ? sen_str_symbol : sen_str_alpha;
+        break;
       case 5 :
-	*d = (c <= 'Z') ? c + 0x20 : c;
-	ctype = (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
-	break;
+        *d = (c <= 'Z') ? c + 0x20 : c;
+        ctype = (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
+        break;
       case 6 :
-	*d = c;
-	ctype = (c == 0x60) ? sen_str_symbol : sen_str_alpha;
-	break;
+        *d = c;
+        ctype = (c == 0x60) ? sen_str_symbol : sen_str_alpha;
+        break;
       case 7 :
-	*d = c;
-	ctype = (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
-	break;
+        *d = c;
+        ctype = (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
+        break;
       default :
-	*d = c;
-	ctype = sen_str_others;
-	break;
+        *d = c;
+        ctype = sen_str_others;
+        break;
       }
     }
     d++;
@@ -841,12 +861,15 @@ normalize_none(sen_nstr *nstr)
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
     if (!(nstr->checks = SEN_MALLOC(size + 1))) {
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
   ch = nstr->checks;
   if (nstr->flags & SEN_STR_WITH_CTYPES) {
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
+      SEN_FREE(nstr->checks);
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
@@ -861,16 +884,16 @@ normalize_none(sen_nstr *nstr)
       continue;
     case 2 :
       if (c == 0x20) {
-	if (removeblankp) {
-	  if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-	  continue;
-	} else {
-	  *d = ' ';
-	  ctype = SEN_NSTR_BLANK|sen_str_symbol;
-	}
+        if (removeblankp) {
+          if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+          continue;
+        } else {
+          *d = ' ';
+          ctype = SEN_NSTR_BLANK|sen_str_symbol;
+        }
       } else {
-	*d = c;
-	ctype = sen_str_symbol;
+        *d = c;
+        ctype = sen_str_symbol;
       }
       break;
     case 3 :
@@ -929,12 +952,15 @@ normalize_latin1(sen_nstr *nstr)
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
     if (!(nstr->checks = SEN_MALLOC(size + 1))) {
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
   ch = nstr->checks;
   if (nstr->flags & SEN_STR_WITH_CTYPES) {
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
+      SEN_FREE(nstr->checks);
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
@@ -949,16 +975,16 @@ normalize_latin1(sen_nstr *nstr)
       continue;
     case 2 :
       if (c == 0x20) {
-	if (removeblankp) {
-	  if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-	  continue;
-	} else {
-	  *d = ' ';
-	  ctype = SEN_NSTR_BLANK|sen_str_symbol;
-	}
+        if (removeblankp) {
+          if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+          continue;
+        } else {
+          *d = ' ';
+          ctype = SEN_NSTR_BLANK|sen_str_symbol;
+        }
       } else {
-	*d = c;
-	ctype = sen_str_symbol;
+        *d = c;
+        ctype = sen_str_symbol;
       }
       break;
     case 3 :
@@ -983,20 +1009,20 @@ normalize_latin1(sen_nstr *nstr)
       break;
     case 8 :
       if (c == 0x8a || c == 0x8c || c == 0x8e) {
-	*d = c + 0x10;
-	ctype = sen_str_alpha;
+        *d = c + 0x10;
+        ctype = sen_str_alpha;
       } else {
-	*d = c;
-	ctype = sen_str_symbol;
+        *d = c;
+        ctype = sen_str_symbol;
       }
       break;
     case 9 :
       if (c == 0x9a || c == 0x9c || c == 0x9e || c == 0x9f) {
-	*d = (c == 0x9f) ? c + 0x60 : c;
-	ctype = sen_str_alpha;
+        *d = (c == 0x9f) ? c + 0x60 : c;
+        ctype = sen_str_alpha;
       } else {
-	*d = c;
-	ctype = sen_str_symbol;
+        *d = c;
+        ctype = sen_str_symbol;
       }
       break;
     case 0x0c :
@@ -1050,12 +1076,15 @@ normalize_koi8r(sen_nstr *nstr)
   d0 = (unsigned char *) nstr->norm;
   if (nstr->flags & SEN_STR_WITH_CHECKS) {
     if (!(nstr->checks = SEN_MALLOC(size + 1))) {
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
   ch = nstr->checks;
   if (nstr->flags & SEN_STR_WITH_CTYPES) {
     if (!(nstr->ctypes = SEN_MALLOC(size + 1))) {
+      SEN_FREE(nstr->checks);
+      SEN_FREE(nstr->norm);
       return sen_memory_exhausted;
     }
   }
@@ -1070,16 +1099,16 @@ normalize_koi8r(sen_nstr *nstr)
       continue;
     case 2 :
       if (c == 0x20) {
-	if (removeblankp) {
-	  if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
-	  continue;
-	} else {
-	  *d = ' ';
-	  ctype = SEN_NSTR_BLANK|sen_str_symbol;
-	}
+        if (removeblankp) {
+          if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
+          continue;
+        } else {
+          *d = ' ';
+          ctype = SEN_NSTR_BLANK|sen_str_symbol;
+        }
       } else {
-	*d = c;
-	ctype = sen_str_symbol;
+        *d = c;
+        ctype = sen_str_symbol;
       }
       break;
     case 3 :
@@ -1108,11 +1137,11 @@ normalize_koi8r(sen_nstr *nstr)
       break;
     case 0x0b :
       if (c == 0xb3) {
-	*d = c - 0x10;
-	ctype = sen_str_alpha;
+        *d = c - 0x10;
+        ctype = sen_str_alpha;
       } else {
-	*d = c;
-	ctype = sen_str_others;
+        *d = c;
+        ctype = sen_str_others;
       }
       break;
     case 0x0c :
@@ -1223,12 +1252,12 @@ sen_str_charlen(const char *str, sen_encoding encoding)
     if (*p & 0x80) {
       int b, w; 
       for (b = 0x40, w = 0; b && (*p & b); b >>= 1, w++);
-      if (!w) { sen_log("invalid utf8 string"); }
+      if (!w) { SEN_LOG(sen_log_warning, "invalid utf8 string"); }
       for (size = 1; w--; size++) {
-	if (!*++p || (*p & 0xc0) != 0x80) {
-	  sen_log("invalid utf8 string");
-	  break;
-	}
+        if (!*++p || (*p & 0xc0) != 0x80) {
+          SEN_LOG(sen_log_warning, "invalid utf8 string");
+          break;
+        }
       }
     } else {
       size = 1;
@@ -1237,9 +1266,9 @@ sen_str_charlen(const char *str, sen_encoding encoding)
   case sen_enc_sjis :
     if (*p & 0x80) {
       if ((0xa1 <= *p && *p <= 0xdf) || !(*(p + 1))) {
-	size = 1;
+        size = 1;
       } else {
-	size = 2;
+        size = 2;
       }
     } else {
       size = 1;
