@@ -85,6 +85,7 @@ void *sen_calloc(size_t size, const char* file, int line);
 void sen_free(void *ptr, const char* file, int line);
 void *sen_realloc(void *ptr, size_t size, const char* file, int line);
 char * sen_strdup(const char *s, const char* file, int line);
+void sen_assert(int cond, const char* file, int line, const char* func);
 int sen_isspace(const char *s, sen_encoding encoding);
 
 #define SEN_MALLOC(s) sen_malloc(s,__FILE__,__LINE__)
@@ -92,6 +93,11 @@ int sen_isspace(const char *s, sen_encoding encoding);
 #define SEN_REALLOC(p,s) sen_realloc(p,s,__FILE__,__LINE__)
 #define SEN_FREE(p) sen_free(p,__FILE__,__LINE__)
 #define SEN_STRDUP(s) sen_strdup(s,__FILE__,__LINE__)
+#ifdef DEBUG
+#define SEN_ASSERT(s) sen_assert((s),__FILE__,__LINE__,__FUNCTION__)
+#else
+#define SEN_ASSERT(s) 
+#endif
 
 #ifdef __cplusplus
 }
