@@ -4,12 +4,12 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
@@ -239,17 +239,18 @@ sis_collect(sen_sym *sym, sen_set *h, sen_id id, uint32_t level)
 
 /* key operation */
 
-inline static char *
-key_at(sen_sym *sym, uint32_t pos)
-{
-  sen_sym_seginfo *si;
-  si = &sym->keyarray[pos >> W_OF_KEY_IN_A_SEGMENT];
-  if (!si->addr) {
-    if (si->segno == -1) { load_all_segments(sym); }
-    SEG_LOAD(sym->io, si);
-    if (!si->addr) { return NULL; }
-  }
-  return (char *)(((byte *)si->addr) + (pos & KEY_MASK_IN_A_SEGMENT));
+#define KEY_AT(sym,pos,ptr) \
+{\
+  sen_sym_seginfo *si = &sym->keyarray[pos >> W_OF_KEY_IN_A_SEGMENT];\
+  if (si->addr) {\
+    ptr = (char *)(((byte *)si->addr) + (pos & KEY_MASK_IN_A_SEGMENT));\
+  } else {\
+    if (si->segno == -1) { load_all_segments(sym); }\
+    SEG_LOAD(sym->io, si);\
+    ptr = si->addr\
+      ? (char *)(((byte *)si->addr) + (pos & KEY_MASK_IN_A_SEGMENT))\
+      : NULL;\
+  }\
 }
 
 inline static uint32_t
@@ -263,7 +264,12 @@ key_put(sen_sym *sym, const uint8_t *key, int len)
     if (segment_new(sym, segment_key)) { return 0; }
     res = sym->header->curr_key;
   }
-  memcpy(key_at(sym, res), key, len);
+  {
+    char *dest;
+    KEY_AT(sym, res, dest);
+    if (!dest) { return 0; }
+    memcpy(dest, key, len);
+  }
   sym->header->curr_key += len;
   return res;
 }
@@ -271,7 +277,13 @@ key_put(sen_sym *sym, const uint8_t *key, int len)
 inline static char *
 pat_node_get_key(sen_sym *sym, pat_node *n)
 {
-  return PAT_IMD(n) ? (char *)&n->key : key_at(sym, n->key);
+  if (PAT_IMD(n)) {
+    return (char *)&n->key;
+  } else {
+    char *res;
+    KEY_AT(sym, n->key, res);
+    return res;
+  }
 }
 
 inline static sen_rc
@@ -290,13 +302,16 @@ pat_node_set_key(sen_sym *sym, pat_node *n, const uint8_t *key, unsigned int len
 
 /* delinfo operation */
 
-#define DL_EMPTY  0
-#define DL_PHASE1 1
-#define DL_PHASE2 2
-#define DL_LD(x) ((x)->bitfield >> 2)
+#define DL_EMPTY  0x00000000
+#define DL_PHASE1 0x00000001
+#define DL_PHASE2 0x00000002
+#define DL_SHARED 0x80000000
+#define DL_LD(x) (((x)->bitfield >> 2) & SEN_SYM_MAX_ID)
 #define DL_ST(x) ((x)->bitfield & 3)
+#define DL_SH(x) ((x)->bitfield & DL_SHARED)
 #define SET_DL_LD(x,v) ((x)->bitfield = DL_ST(x) | ((v) << 2))
 #define SET_DL_ST(x,v) ((x)->bitfield = ((x)->bitfield & 0xfffffffc) | ((v) & 3))
+#define SET_DL_SH(x,v) ((x)->bitfield = ((x)->bitfield & ~DL_SHARED) | (v))
 
 inline static syn_sym_delinfo *
 delinfo_search(sen_sym *sym, sen_id id)
@@ -339,12 +354,11 @@ delinfo_turn_2(sen_sym *sym, syn_sym_delinfo *di)
     p0 = &rn->r;
     while ((r = *p0)) {
       if (r == d) {
-	p = p0;
-	break;
+        p = p0;
+        break;
       }
       if (!(rn = pat_at(sym, r))) { return sen_invalid_format; }
-      if (len <= rn->check) { return sen_invalid_format; }
-      if (c >= rn->check) { return sen_invalid_format; }
+      if (rn->check <= c || len <= rn->check) { break; }
       c = rn->check;
       p0 = nth_bit((uint8_t *)key, c) ? &rn->r : &rn->l;
     }
@@ -359,16 +373,16 @@ delinfo_turn_2(sen_sym *sym, syn_sym_delinfo *di)
     int j;
     sen_id dd;
     syn_sym_delinfo *ddi;
-    sen_log("failed to find d=%d", d);
+    SEN_LOG(sen_log_debug, "failed to find d=%d", d);
     for (j = (sym->header->curr_del2 + 1) & SEN_SYM_MDELINFOS;
-	 j != sym->header->curr_del;
-	 j = (j + 1) & SEN_SYM_MDELINFOS) {
+         j != sym->header->curr_del;
+         j = (j + 1) & SEN_SYM_MDELINFOS) {
       ddi = &sym->header->delinfos[j];
       if (DL_ST(ddi) != DL_PHASE1) { continue; }
       if (!(ln = pat_at(sym, DL_LD(ddi)))) { continue; }
       if (!(dd = ddi->d)) { continue; }
       if (d == DL_LD(ddi)) {
-	sen_log("found!!!, d(%d) become ld of (%d)", d, dd);
+        SEN_LOG(sen_log_debug, "found!!!, d(%d) become ld of (%d)", d, dd);
       }
     }
     /* debug */
@@ -384,14 +398,24 @@ delinfo_turn_3(sen_sym *sym, syn_sym_delinfo *di)
 {
   pat_node *dn;
   uint32_t size;
-  const char *key;
   if (DL_ST(di) != DL_PHASE2) { return sen_success; }
   if (!(dn = pat_at(sym, di->d))) { return sen_invalid_argument; }
-  if (!(key = pat_node_get_key(sym, dn))) { return sen_invalid_argument; }
-  if (sym->key_size) {
-    size = sym->key_size;
+  if (DL_SH(di)) {
+    SET_PAT_IMD(dn, PAT_IMMEDIATE);
+    size = 0;
   } else {
-    size = dn->check = (uint16_t)(strlen(key) + 1);
+    if (sym->key_size) {
+      size = sym->key_size;
+    } else {
+      // size = dn->check = (uint16_t)(strlen(key) + 1);
+      if (PAT_IMD(dn)) {
+        size = 0;
+      } else {
+        const char *key = pat_node_get_key(sym, dn);
+        if (!key) { return sen_invalid_argument; }
+        size = strlen(key) + 1;
+      }
+    }
   }
   SET_DL_ST(di, DL_EMPTY);
   dn->r = SEN_SYM_DELETED;
@@ -409,13 +433,13 @@ delinfo_new(sen_sym *sym)
             - (SEN_SYM_NDELINFOS / 2);
   while (gap-- > 0) {
     if (delinfo_turn_2(sym, &sym->header->delinfos[sym->header->curr_del2])) {
-      sen_log("d2 failed: %d", DL_LD(&sym->header->delinfos[sym->header->curr_del2]));
+      SEN_LOG(sen_log_crit, "d2 failed: %d", DL_LD(&sym->header->delinfos[sym->header->curr_del2]));
     }
     sym->header->curr_del2 = (sym->header->curr_del2 + 1) & SEN_SYM_MDELINFOS;
   }
   if (n == sym->header->curr_del3) {
     if (delinfo_turn_3(sym, &sym->header->delinfos[sym->header->curr_del3])) {
-      sen_log("d3 failed: %d", DL_LD(&sym->header->delinfos[sym->header->curr_del3]));
+      SEN_LOG(sen_log_crit, "d3 failed: %d", DL_LD(&sym->header->delinfos[sym->header->curr_del3]));
     }
     sym->header->curr_del3 = (sym->header->curr_del3 + 1) & SEN_SYM_MDELINFOS;
   }
@@ -427,7 +451,7 @@ delinfo_new(sen_sym *sym)
 
 sen_sym *
 sen_sym_create(const char *path, uint32_t key_size,
-	       uint32_t flags, sen_encoding encoding)
+               uint32_t flags, sen_encoding encoding)
 {
   int i;
   sen_io *io;
@@ -435,7 +459,7 @@ sen_sym_create(const char *path, uint32_t key_size,
   pat_node *node0;
   struct sen_sym_header *header;
   io = sen_io_create(path, SEN_SYM_HEADER_SIZE, SEN_SYM_SEGMENT_SIZE,
-		     SEN_SYM_MAX_SEGMENT, sen_io_auto, SEN_SYM_MAX_SEGMENT);
+                     SEN_SYM_MAX_SEGMENT, sen_io_auto, SEN_SYM_MAX_SEGMENT);
   if (!io) { return NULL; }
   header = sen_io_header(io);
   memcpy(header->idstr, SEN_IDSTR, 16);
@@ -518,7 +542,9 @@ sen_sym_open(const char *path)
 sen_rc
 sen_sym_close(sen_sym *sym)
 {
-  sen_rc rc = sen_io_close(sym->io);
+  sen_rc rc;
+  if (!sym) { return sen_invalid_argument; }
+  rc = sen_io_close(sym->io);
   SEN_FREE(sym);
   return rc;
 }
@@ -532,7 +558,7 @@ sen_sym_remove(const char *path)
 }
 
 inline static sen_id
-_sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new)
+_sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new, uint32_t *lkey)
 {
   sen_id r, r0, *p0, *p1 = NULL;
   pat_node *rn, *rn0;
@@ -550,51 +576,68 @@ _sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new)
     while ((r0 = *p0)) {
       if (!(rn0 = pat_at(sym, r0))) { return 0; }
       if (c0 < rn0->check && rn0->check < len) {
-	c1 = c0; c0 = rn0->check;
-	p1 = p0; p0 = nth_bit(key, c0) ? &rn0->r : &rn0->l;
+        c1 = c0; c0 = rn0->check;
+        p1 = p0; p0 = nth_bit(key, c0) ? &rn0->r : &rn0->l;
       } else {
-	if (rn0->check < len && !memcmp(pat_node_get_key(sym, rn0), key, size)) {
-	  return r0;
-	}
-	break;
+        if (rn0->check < len && !memcmp(pat_node_get_key(sym, rn0), key, size)) {
+          return r0;
+        }
+        break;
       }
     }
     for (c = 0, s = pat_node_get_key(sym, rn0), d = (char *)key; *s == *d; c += 8, s++, d++);
     for (xor = *s ^ *d, mask = 0x80; !(xor & mask); mask >>= 1, c++);
     // for test
     if (r0 && r0 != sen_sym_at(sym, pat_node_get_key(sym, rn0))) {
-      sen_log("deleted node is used as ld %d(%d:%s) %d", r0, PAT_DEL(rn0), pat_node_get_key(sym, rn0), sen_sym_at(sym, pat_node_get_key(sym, rn0)));
+      SEN_LOG(sen_log_debug, "deleted node is used as ld %d(%d:%s) %d", r0, PAT_DEL(rn0), pat_node_get_key(sym, rn0), sen_sym_at(sym, pat_node_get_key(sym, rn0)));
     }
 
     if (c == c0 && !*p0) {
       if (c < len - 1) { c++; }
     } else {
       if (c < c0) {
-	if (c > c1) {
-	  p0 = p1;
-	} else {
-	  rn0 = pat_at(sym, 0); p0 = &rn0->r;
-	  while ((r0 = *p0)) {
-	    if (!(rn0 = pat_at(sym, r0))) { return 0; }
-	    if (c < rn0->check) { break; }
-	    p0 = nth_bit(key, rn0->check) ? &rn0->r : &rn0->l;
-	  }
-	}
+        if (c > c1) {
+          p0 = p1;
+        } else {
+          rn0 = pat_at(sym, 0); p0 = &rn0->r;
+          while ((r0 = *p0)) {
+            if (!(rn0 = pat_at(sym, r0))) { return 0; }
+            if (c < rn0->check) { break; }
+            p0 = nth_bit(key, rn0->check) ? &rn0->r : &rn0->l;
+          }
+        }
       }
     }
     if (c >= len) { return 0; }
   } else {
     c = (len - 1 > SEN_SYM_MAX_KEY_LENGTH) ? SEN_SYM_MAX_KEY_LENGTH : len - 1;
   }
-  if (sym->header->garbages[size]) {
-    r = sym->header->garbages[size];
-    sym->header->nrecords++;
-    rn = pat_at(sym, r);
-    sym->header->garbages[size] = rn->l;
-    memcpy(pat_node_get_key(sym, rn), key, size);
-  } else {
-    if (!(rn = pat_node_new(sym, &r))) { return 0; }
-    pat_node_set_key(sym, rn, key, (unsigned int)size);
+  {
+    size_t size2 = size > sizeof(uint32_t) ? size : 0;
+    if (*lkey && size2) {
+      if (sym->header->garbages[0]) {
+        r = sym->header->garbages[0];
+        sym->header->nrecords++;
+        rn = pat_at(sym, r);
+        sym->header->garbages[0] = rn->l;
+      } else {
+        if (!(rn = pat_node_new(sym, &r))) { return 0; }
+      }
+      SET_PAT_IMD(rn, 0);
+      rn->key = *lkey;
+    } else {
+      if (sym->header->garbages[size2]) {
+        r = sym->header->garbages[size2];
+        sym->header->nrecords++;
+        rn = pat_at(sym, r);
+        sym->header->garbages[size2] = rn->l;
+        memcpy(pat_node_get_key(sym, rn), key, size);
+      } else {
+        if (!(rn = pat_node_new(sym, &r))) { return 0; }
+        pat_node_set_key(sym, rn, key, (unsigned int)size);
+      }
+      *lkey = rn->key;
+    }
   }
   rn->check = c;
   SET_PAT_DEL(rn, 0);
@@ -612,10 +655,11 @@ _sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new)
 }
 
 inline static int
-chop(sen_sym *sym, const char **key)
+chop(sen_sym *sym, const char **key, uint32_t *lkey)
 {
   size_t len = sen_str_charlen(*key, sym->encoding);
   if (len) {
+    *lkey += len;
     *key += len;
     return **key;
   } else {
@@ -624,11 +668,11 @@ chop(sen_sym *sym, const char **key)
 }
 
 sen_id
-sen_sym_get(sen_sym *sym, const char *key)
+sen_sym_get(sen_sym *sym, const void *key)
 {
-  uint32_t new;
+  uint32_t new, lkey = 0;
   sen_id r0;
-  r0 = _sen_sym_get(sym, (uint8_t *)key, &new);
+  r0 = _sen_sym_get(sym, (uint8_t *)key, &new, &lkey);
   if ((sym->flags & SEN_SYM_WITH_SIS) && (*((uint8_t *)key) & 0x80)) { // todo: refine!!
     sis_node *sl, *sr;
     sen_id l = r0, r;
@@ -636,20 +680,21 @@ sen_sym_get(sen_sym *sym, const char *key)
       const char *sis = key;
       sl->children = l;
       sl->sibling = 0;
-      while (chop(sym, &sis)) {
-	r = _sen_sym_get(sym, (uint8_t *)sis, &new);
-	if (!(sr = sis_get(sym, r))) { break; }
-	if (new) {
-	  sl->sibling = r;
-	  sr->children = l;
-	  sr->sibling = 0;
-	} else {
-	  sl->sibling = sr->children;
-	  sr->children = l;
-	  break;
-	}
-	l = r;
-	sl = sr;
+      while (chop(sym, &sis, &lkey)) {
+        if (!(*sis & 0x80)) { break; }
+        r = _sen_sym_get(sym, (uint8_t *)sis, &new, &lkey);
+        if (!(sr = sis_get(sym, r))) { break; }
+        if (new) {
+          sl->sibling = r;
+          sr->children = l;
+          sr->sibling = 0;
+        } else {
+          sl->sibling = sr->children;
+          sr->children = l;
+          break;
+        }
+        l = r;
+        sl = sr;
       }
     }
   }
@@ -657,7 +702,7 @@ sen_sym_get(sen_sym *sym, const char *key)
 }
 
 sen_id
-sen_sym_at(sen_sym *sym, const char *key)
+sen_sym_at(sen_sym *sym, const void *key)
 {
   sen_id r;
   pat_node *rn;
@@ -703,7 +748,7 @@ get_tc(sen_sym *sym, sen_set *h, pat_node *rn, unsigned int len)
 }
 
 sen_set *
-sen_sym_prefix_search(sen_sym *sym, const char *key)
+sen_sym_prefix_search(sen_sym *sym, const void *key)
 {
   int c;
   size_t len;
@@ -723,6 +768,8 @@ sen_sym_prefix_search(sen_sym *sym, const char *key)
     }
     if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
       sen_set *h = sen_set_open(sizeof(sen_id), 0, 0);
+      SEN_ASSERT(h);
+//      if (!h) { return NULL; }
       sen_set_get(h, &r, NULL);
       if (rn->check >= len) { get_tc(sym, h, rn, (unsigned int)len); }
       return h;
@@ -733,13 +780,15 @@ sen_sym_prefix_search(sen_sym *sym, const char *key)
 }
 
 sen_set *
-sen_sym_suffix_search(sen_sym *sym, const char *key)
+sen_sym_suffix_search(sen_sym *sym, const void *key)
 {
   sen_id r;
   uint32_t *offset;
   if (!key || sym->key_size) { return NULL; }
   if ((r = sen_sym_at(sym, key))) {
     sen_set *h = sen_set_open(sizeof(sen_id), sizeof(uint32_t), 0);
+    SEN_ASSERT(h);
+//    if (!h) { return NULL; }
     sen_set_get(h, &r, (void **) &offset);
     *offset = 0;
     if (sym->flags & SEN_SYM_WITH_SIS) { sis_collect(sym, h, r, 1); }
@@ -749,7 +798,7 @@ sen_sym_suffix_search(sen_sym *sym, const char *key)
 }
 
 sen_id
-sen_sym_common_prefix_search(sen_sym *sym, const char *key)
+sen_sym_common_prefix_search(sen_sym *sym, const void *key)
 {
   pat_node *rn, *rnl;
   sen_id r, rl, r2 = SEN_SYM_NIL;
@@ -767,20 +816,20 @@ sen_sym_common_prefix_search(sen_sym *sym, const char *key)
     }
     c = rn->check;
     n = c >> 3;
-    while (len < n) { if (!key[len++]) { goto exit; } }
+    while (len < n) { if (!((char *)key)[len++]) { goto exit; } }
     nth_bit = ((uint8_t *)key)[n] & (0x80 >> (c & 7));
     if (nth_bit) {
       if (len2 <= c) {
-	len2 = (n + 1) * 8;
-	for (cl = c, rl = rn->l; rl; cl = rnl->check, rl = rnl->l) {
-	  if (!(rnl = pat_at(sym, rl))) { break; /* corrupt? */ }
-	  if (len2 <= rnl->check) { break; }
-	  if (rnl->check <= cl) {
-	    char *p = pat_node_get_key(sym, rnl);
-	    if (!p[n] && !memcmp(p, key, n)) { r2 = rl; }
-	    break;
-	  }
-	}
+        len2 = (n + 1) * 8;
+        for (cl = c, rl = rn->l; rl; cl = rnl->check, rl = rnl->l) {
+          if (!(rnl = pat_at(sym, rl))) { break; /* corrupt? */ }
+          if (len2 <= rnl->check) { break; }
+          if (rnl->check <= cl) {
+            char *p = pat_node_get_key(sym, rnl);
+            if (!p[n] && !memcmp(p, key, n)) { r2 = rl; }
+            break;
+          }
+        }
       }
       r = rn->r;
     } else {
@@ -791,8 +840,8 @@ exit :
   return r2;
 }
 
-sen_rc
-sen_sym_del(sen_sym *sym, const char *key)
+inline static sen_rc
+_sen_sym_del(sen_sym *sym, const char *key, int shared)
 {
   syn_sym_delinfo *di;
   uint8_t direction;
@@ -803,6 +852,7 @@ sen_sym_del(sen_sym *sym, const char *key)
   if (!key) { return sen_invalid_argument; }
   if (!len) { len = (strlen(key) + 1) * 8; }
   di = delinfo_new(sym); /* must be called before find rn */
+  SET_DL_SH(di, shared ? DL_SHARED : 0);
   rn = pat_at(sym, 0);
   c = -1;
   p = &rn->r;
@@ -812,8 +862,8 @@ sen_sym_del(sen_sym *sym, const char *key)
     if (len <= rn->check) { return sen_invalid_argument; }
     if (c >= rn->check) {
       if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
-	break; /* found */
-      } else { 	return sen_invalid_argument; }
+        break; /* found */
+      } else {  return sen_invalid_argument; }
     }
     c0 = c; c = rn->check;
     p0 = p; p = nth_bit((uint8_t *)key, c) ? &rn->r : &rn->l;
@@ -827,10 +877,10 @@ sen_sym_del(sen_sym *sym, const char *key)
     if (otherside) {
       rno = pat_at(sym, otherside);
       if (rno && c0 < rno->check && rno->check <= c) {
-	if (!delinfo_search(sym, otherside)) {
-	  sen_log("caution: no delinfo found %d", otherside);
-	}
-	rno->check = 0;
+        if (!delinfo_search(sym, otherside)) {
+          SEN_LOG(sen_log_error, "no delinfo found %d", otherside);
+        }
+        rno->check = 0;
       }
     }
     *p0 = otherside;
@@ -842,45 +892,58 @@ sen_sym_del(sen_sym *sym, const char *key)
       SET_PAT_DEL(rn, 0);
       SET_DL_ST(di, DL_PHASE2);
       if (ddi) {
-	SET_PAT_DEL(rn0, 0);
-	SET_DL_ST(ddi, DL_PHASE2);
-	if (ddi == ldi) {
-	  di->d = DL_LD(ddi);
-	} else {
-	  SET_DL_LD(ldi, DL_LD(ddi));
-	  di->d = r;
-	}
+        SET_PAT_DEL(rn0, 0);
+        SET_DL_ST(ddi, DL_PHASE2);
+        if (ddi == ldi) {
+          if (r != DL_LD(ddi)) {
+            SEN_LOG(sen_log_error, "r(%d) != DL_LD(ddi)(%d)", r, DL_LD(ddi));
+          }
+          di->d = r;
+        } else {
+          SET_DL_LD(ldi, DL_LD(ddi));
+          di->d = r;
+        }
       } else {
-	SET_PAT_DEL(rn0, PAT_DELETING);
-	SET_DL_LD(ldi, *p0);
-	di->d = r;
+        SET_PAT_DEL(rn0, PAT_DELETING);
+        SET_DL_LD(ldi, *p0);
+        di->d = r;
       }
     } else {
       SET_PAT_DEL(rn, PAT_DELETING);
       if (ddi) {
-	SET_PAT_DEL(rn0, 0);
-	ddi->d = r;
-	SET_DL_ST(di, DL_PHASE2);
-	di->d = *p0;
+        if (ddi->d != *p0) {
+          SEN_LOG(sen_log_error, "ddi->d(%d) != *p0(%d)", ddi->d, *p0);
+        }
+        SET_PAT_DEL(rn0, 0);
+        SET_DL_ST(ddi, DL_PHASE2);
+        SET_DL_ST(di, DL_PHASE1);
+        SET_DL_LD(di, DL_LD(ddi));
+        di->d = r;
+        /*
+        SET_PAT_DEL(rn0, 0);
+        ddi->d = r;
+        SET_DL_ST(di, DL_PHASE2);
+        di->d = *p0;
+        */
       } else {
-	SET_PAT_DEL(rn0, PAT_DELETING);
-	SET_DL_ST(di, DL_PHASE1);
-	SET_DL_LD(di, *p0);
-	di->d = r;
-	// sen_log("sym_del d=%d ld=%d stat=%d", r, *p0, DL_PHASE1);
+        SET_PAT_DEL(rn0, PAT_DELETING);
+        SET_DL_ST(di, DL_PHASE1);
+        SET_DL_LD(di, *p0);
+        di->d = r;
+        // sen_log("sym_del d=%d ld=%d stat=%d", r, *p0, DL_PHASE1);
       }
     }
     if (*p0 == otherside) {
       rn0->check = 0;
     } else {
       if (otherside) {
-	rno = pat_at(sym, otherside);
-	if (rno && c0 < rno->check && rno->check <= c) {
-	  if (!delinfo_search(sym, otherside)) {
-	    sen_log("caution! no delinfo found %d", otherside);
-	  }
-	  rno->check = 0;
-	}
+        rno = pat_at(sym, otherside);
+        if (rno && c0 < rno->check && rno->check <= c) {
+          if (!delinfo_search(sym, otherside)) {
+            SEN_LOG(sen_log_error, "no delinfo found %d", otherside);
+          }
+          rno->check = 0;
+        }
       }
       *p0 = otherside;
     }
@@ -889,6 +952,12 @@ sen_sym_del(sen_sym *sym, const char *key)
   return sen_success;
 }
 
+sen_rc
+sen_sym_del(sen_sym *sym, const void *key)
+{
+  return _sen_sym_del(sym, key, 0);
+}
+
 uint32_t
 sen_sym_size(sen_sym *sym)
 {
@@ -935,7 +1004,7 @@ sen_sym_pocket_set(sen_sym *sym, sen_id id, unsigned int value)
 
 sen_rc
 sen_sym_info(sen_sym *sym, int *key_size, unsigned *flags,
-	     sen_encoding *encoding, unsigned *nrecords, unsigned *file_size)
+             sen_encoding *encoding, unsigned *nrecords, unsigned *file_size)
 {
   if (!sym) { return sen_invalid_argument; }
   if (key_size) { *key_size = sym->key_size; }
@@ -953,30 +1022,66 @@ sen_sym_info(sen_sym *sym, int *key_size, unsigned *flags,
   return sen_success;
 }
 
-sen_id
-sen_sym_del_with_sis(sen_sym *sym, sen_id id)
+int
+sen_sym_del_with_sis(sen_sym *sym, sen_id id,
+                     int(*func)(sen_id, void *), void *func_arg)
 {
-  sen_id *p, sid = SEN_SYM_NIL;
-  const char *key = _sen_sym_key(sym, id);
+  int level = 0, shared;
+  const char *key = NULL, *_key;
   sis_node *sp, *ss = NULL, *si = sis_at(sym, id);
-  if (key && si) {
-    if (si->children && si->children != id) { return SEN_SYM_NIL; }
-    sen_sym_del(sym, key);
-    if ((*key & 0x80) && chop(sym, &key)) {
-      if ((sid = sen_sym_at(sym, key)) && (ss = sis_at(sym, sid))) {
-	for (p = &ss->children; *p && *p != sid; p = &sp->sibling) {
-	  if (*p == id) {
-	    *p = si->sibling;
-	    break;
-	  }
-	  sp = sis_at(sym, *p);
-	}
-      }
+  while (id) {
+    if ((si && si->children && si->children != id) || !func(id, func_arg)) { break; }
+    if (!(_key = _sen_sym_key(sym, id))) { break; }
+    if (_key == key) {
+      shared = 1;
+    } else {
+      key = _key;
+      shared = 0;
     }
-    si->sibling = 0;
-    si->children = 0;
+    _sen_sym_del(sym, key, shared);
+    if (si) {
+      sen_id *p, sid;
+      uint32_t lkey = 0;
+      if ((*key & 0x80) && chop(sym, &key, &lkey)) {
+        if ((sid = sen_sym_at(sym, key)) && (ss = sis_at(sym, sid))) {
+          for (p = &ss->children; *p && *p != sid; p = &sp->sibling) {
+            if (*p == id) {
+              *p = si->sibling;
+              break;
+            }
+            sp = sis_at(sym, *p);
+          }
+        }
+      } else {
+        sid = SEN_SYM_NIL;
+      }
+      si->sibling = 0;
+      si->children = 0;
+      id = sid;
+      si = ss;
+    } else {
+      id = SEN_SYM_NIL;
+    }
+    level++;
   }
-  return (ss && (!ss->children || ss->children == sid)) ? sid : SEN_SYM_NIL;
+  if (level) {
+    uint32_t lkey = 0;
+    while (id && key) {
+      if (_sen_sym_key(sym, id) != key) { break; }
+      {
+        pat_node *rn = pat_at(sym, id);
+        if (lkey) {
+          rn->key = lkey;
+        } else {
+          pat_node_set_key(sym, rn, (uint8_t *)key, (unsigned int)(strlen(key) + 1));
+          lkey = rn->key;
+        }
+      }
+      if (!((*key & 0x80) && chop(sym, &key, &lkey))) { break; }
+      id = sen_sym_at(sym, key);
+    }
+  }
+  return level;
 }
 
 sen_id
@@ -999,7 +1104,7 @@ sen_sym_dump(sen_sym *sym, int to)
   if (to < 0) { to = sym->header->curr_rec; }
   for (i = 0; i <= (unsigned int)to; i++) {
     rn = pat_at(sym, i);
-    sen_log("%d(%p):[%d,%d](%d)\"%s\"",
+    SEN_LOG(sen_log_debug, "%d(%p):[%d,%d](%d)\"%s\"",
             i, rn, rn->r, rn->l, rn->check, pat_node_get_key(sym, rn));
   }
 }
@@ -1013,16 +1118,17 @@ sen_sym_inspect_key(sen_sym *sym, const uint8_t *key)
   size_t len = sym->key_size * 8;
   if (!key) { return SEN_SYM_NIL; }
   if (!len) { len = (strlen((char *)key) + 1) * 8; }
-  sen_log("sen_sym_inspect_key key=%s len=%d", key, len);
+  SEN_LOG(sen_log_debug, "sen_sym_inspect_key key=%s len=%d", key, len);
   for (r = pat_at(sym, 0)->r; r; r = nth_bit(key, c) ? rn->r : rn->l) {
-    if (!(rn = pat_at(sym, r))) { sen_log("corrupt?"); break; /* corrupt? */ }
-    sen_log("%d[%d,%d] key='%s' c=%d rn->c=%d len=%d", r, rn->r, rn->l, pat_node_get_key(sym, rn), c, rn->check, len);
+    if (!(rn = pat_at(sym, r))) { SEN_LOG(sen_log_debug, "corrupt?"); break; /* corrupt? */ }
+    SEN_LOG(sen_log_debug,
+            "%d[%d,%d] key='%s' c=%d rn->c=%d len=%d", r, rn->r, rn->l, pat_node_get_key(sym, rn), c, rn->check, len);
     if (len <= rn->check) { break; }
     if (rn->check <= c) {
-      sen_log("c=%d key=%s",c, pat_node_get_key(sym, rn));
+      SEN_LOG(sen_log_debug, "c=%d key=%s",c, pat_node_get_key(sym, rn));
       if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
-	sen_log("found! r=%d", r);
-	return r;
+        SEN_LOG(sen_log_debug, "found! r=%d", r);
+        return r;
       }
       break;
     }
@@ -1036,39 +1142,42 @@ sen_sym_inspect(sen_sym *sym)
 {
   int i;
   printf("idstr=%s\nkey_size=%d\nnrecords=%d\ncurr_rec=%d\ncurr_key=%d\ncurr_del=%d\ncurr_del2=%d\ncurr_del3=%d\n",
-	 sym->header->idstr,
-	 sym->header->key_size,
-	 sym->header->nrecords,
-	 sym->header->curr_rec,
-	 sym->header->curr_key,
-	 sym->header->curr_del,
-	 sym->header->curr_del2,
-	 sym->header->curr_del3);
+         sym->header->idstr,
+         sym->header->key_size,
+         sym->header->nrecords,
+         sym->header->curr_rec,
+         sym->header->curr_key,
+         sym->header->curr_del,
+         sym->header->curr_del2,
+         sym->header->curr_del3);
   {
-    int se = 0, ss = 0, sa = 0, si = 0;
+    int se = 0, sk = 0, sa = 0, ss = 0, si = 0;
     for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
       switch (sym->header->segments[i]) {
       case segment_empty :
-	se++;
-	break;
+        se++;
+        break;
       case segment_key :
-	ss++;
-	break;
+        sk++;
+        break;
       case segment_pat :
-	sa++;
-	break;
+        sa++;
+        break;
+      case segment_sis :
+        ss++;
+        break;
       default :
-	si++;
+        si++;
       }
     }
-    printf("seg_empty=%d, seg_string=%d, seg_array=%d, seg_invalid=%d\n", se, ss, sa, si);
+    printf("seg_empty=%d, seg_string=%d, seg_array=%d, seg_sis=%d, seg_invalid=%d\n", se, sk, sa, ss, si);
   }
   {
     int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
     printf("del2=%d, del=%d\n", sym->header->curr_del2, sym->header->curr_del);
     for (i = sym->header->curr_del2;
-	 i != sym->header->curr_del;
-	 i = (i + 1) & SEN_SYM_MDELINFOS) {
+         i != sym->header->curr_del;
+         i = (i + 1) & SEN_SYM_MDELINFOS) {
       n++;
       switch (DL_ST(&sym->header->delinfos[i])) {
       case 0 : s0++; break;
@@ -1082,8 +1191,8 @@ sen_sym_inspect(sen_sym *sym)
   {
     int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
     for (i = sym->header->curr_del3;
-	 i != sym->header->curr_del2;
-	 i = (i + 1) & SEN_SYM_MDELINFOS) {
+         i != sym->header->curr_del2;
+         i = (i + 1) & SEN_SYM_MDELINFOS) {
       n++;
       switch (DL_ST(&sym->header->delinfos[i])) {
       case 0 : s0++; break;
@@ -1097,8 +1206,8 @@ sen_sym_inspect(sen_sym *sym)
   {
     int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
     for (i = sym->header->curr_del;
-	 i != sym->header->curr_del3;
-	 i = (i + 1) & SEN_SYM_MDELINFOS) {
+         i != sym->header->curr_del3;
+         i = (i + 1) & SEN_SYM_MDELINFOS) {
       n++;
       switch (DL_ST(&sym->header->delinfos[i])) {
       case 0 : s0++; break;
@@ -1116,11 +1225,11 @@ sen_sym_inspect(sen_sym *sym)
     for (i = 1; i <= sym->header->curr_rec; i++) {
       tn = pat_at(sym, i);
       if (tn) {
-	n++;
-	if (tn->r == SEN_SYM_DELETED) { ndel++; }
-	if (PAT_DEL(tn)) { nding++; }
+        n++;
+        if (tn->r == SEN_SYM_DELETED) { ndel++; }
+        if (PAT_DEL(tn)) { nding++; }
       } else {
-	nn++;
+        nn++;
       }
     }
     printf("array: n=%d, nil=%d, del=%d, deleting=%d\n", n, nn, ndel, nding);
