@@ -35,7 +35,7 @@
 #include "io.h"
 #endif /* _SEN_IO_H */
 
-#ifdef	__cplusplus
+#ifdef  __cplusplus
 extern "C" {
 #endif
 
@@ -91,7 +91,9 @@ struct _sen_sym {
 };
 
 const char *_sen_sym_key(sen_sym *sym, sen_id id);
-sen_id sen_sym_del_with_sis(sen_sym *sym, sen_id id);
+// sen_id sen_sym_del_with_sis(sen_sym *sym, sen_id id);
+int sen_sym_del_with_sis(sen_sym *sym, sen_id id,
+                         int(*func)(sen_id, void *), void *func_arg);
 
 #ifdef __cplusplus
 }
