@@ -4,12 +4,12 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
@@ -96,7 +96,9 @@ typedef enum {
   sen_enc_none,
   sen_enc_euc_jp,
   sen_enc_utf8,
-  sen_enc_sjis
+  sen_enc_sjis,
+  sen_enc_latin1,
+  sen_enc_koi8r
 } sen_encoding;
 
 typedef enum {
@@ -259,12 +261,14 @@ sen_rc sen_index_close(sen_index *i);
 sen_rc sen_index_remove(const char *path);
 sen_rc sen_index_rename(const char *old_name, const char *new_name);
 sen_rc sen_index_upd(sen_index *i, const void *key,
-                     const char *oldvalue, const char *newvalue);
-sen_records *sen_index_sel(sen_index *i, const char *string);
-int sen_records_next(sen_records *r, void *keybuf, int bufsize, int *score);
+                     const char *oldvalue, unsigned int oldvalue_len,
+                     const char *newvalue, unsigned int newvalue_len);
+sen_records *sen_index_sel(sen_index *i,
+                           const char *string, unsigned int string_len);
+int sen_records_next(sen_records *r, void *keybuf, int buf_size, int *score);
 sen_rc sen_records_rewind(sen_records *r);
 int sen_records_curr_score(sen_records *r);
-int sen_records_curr_key(sen_records *r, void *keybuf, int bufsize);
+int sen_records_curr_key(sen_records *r, void *keybuf, int buf_size);
 int sen_records_nhits(sen_records *r);
 int sen_records_find(sen_records *r, const void *key);
 sen_rc sen_records_close(sen_records *r);
@@ -273,9 +277,11 @@ sen_rc sen_records_close(sen_records *r);
 
 sen_values *sen_values_open(void);
 sen_rc sen_values_close(sen_values *v);
-sen_rc sen_values_add(sen_values *v, const char *str, unsigned int weight);
+sen_rc sen_values_add(sen_values *v,
+                      const char *str, unsigned int str_len,
+                      unsigned int weight);
 
-sen_records *sen_records_open(sen_rec_unit record_unit, 
+sen_records *sen_records_open(sen_rec_unit record_unit,
                               sen_rec_unit subrec_unit,
                               unsigned int max_n_subrecs);
 sen_records *sen_records_union(sen_records *a, sen_records *b);
@@ -289,31 +295,34 @@ const sen_recordh * sen_records_at(sen_records *r, const void *key,
                                    unsigned section, unsigned pos,
                                    int *score, int *n_subrecs);
 sen_rc sen_record_info(sen_records *r, const sen_recordh *rh,
-                       void *keybuf, int bufsize, int *keysize,
+                       void *keybuf, int buf_size, int *key_size,
                        int *section, int *pos, int *score, int *n_subrecs);
 sen_rc sen_record_subrec_info(sen_records *r, const sen_recordh *rh,
-                              int index, void *keybuf, int bufsize,
-                              int *keysize, int *section, int *pos, int *score);
+                              int index, void *keybuf, int buf_size,
+                              int *key_size, int *section, int *pos, int *score);
 sen_index *sen_index_create_with_keys(const char *path, sen_sym *keys, int flags,
                                       int initial_n_segments, sen_encoding encoding);
 sen_index *sen_index_open_with_keys(const char *path, sen_sym *keys);
 sen_rc sen_index_update(sen_index *i, const void *key, unsigned int section,
                         sen_values *oldvalues, sen_values *newvalues);
-sen_rc sen_index_select(sen_index *i, const char *string, sen_records *r,
+sen_rc sen_index_select(sen_index *i,
+                        const char *string, unsigned int string_len,
+                        sen_records *r,
                         sen_sel_operator op, sen_select_optarg *optarg);
 sen_rc sen_index_info(sen_index *i, int *key_size, int *flags,
                       int *initial_n_segments, sen_encoding *encoding,
                       unsigned *nrecords_keys, unsigned *file_size_keys,
                       unsigned *nrecords_lexicon, unsigned *file_size_lexicon,
                       unsigned *inv_seg_size, unsigned *inv_chunk_size);
-int sen_index_path(sen_index *i, char *pathbuf, int bufsize);
+int sen_index_path(sen_index *i, char *pathbuf, int buf_size);
 /*
 sen_set * sen_index_related_terms(sen_index *index, const char *string,
                                   const char *(*fetcher)(void *, void *),
                                   void *fetcher_arg);
 */
 
-sen_query *sen_query_open(const char *str, sen_sel_operator default_op,
+sen_query *sen_query_open(const char *str, unsigned int str_len,
+                          sen_sel_operator default_op,
                           int max_exprs, sen_encoding encoding);
 const char *sen_query_rest(sen_query *q);
 sen_rc sen_query_close(sen_query *q);
@@ -333,7 +342,7 @@ sen_rc sen_set_del(sen_set *set, sen_set_eh *e);
 sen_set_cursor *sen_set_cursor_open(sen_set *set);
 sen_set_eh *sen_set_cursor_next(sen_set_cursor *cursor, void **key, void **value);
 sen_rc sen_set_cursor_close(sen_set_cursor *cursor);
-sen_rc sen_set_element_info(sen_set *set, const sen_set_eh *e, 
+sen_rc sen_set_element_info(sen_set *set, const sen_set_eh *e,
                             void **key, void **value);
 sen_set *sen_set_union(sen_set *a, sen_set *b);
 sen_set *sen_set_subtract(sen_set *a, sen_set *b);
@@ -352,32 +361,38 @@ sen_rc sen_sym_remove(const char *path);
 /* Lookup the sym table and find the id of the corresponding entry.
  * If no matches are found, create a new entry, and return that ID
  */
-sen_id sen_sym_get(sen_sym *sym, const char *key);
+sen_id sen_sym_get(sen_sym *sym, const void *key);
 
 /* Lookup the sym table and find the id of the corresponding entry.
  * If no matches are found return SEN_SYM_NIL
  */
-sen_id sen_sym_at(sen_sym *sym, const char *key);
-sen_rc sen_sym_del(sen_sym *sym, const char *key);
+sen_id sen_sym_at(sen_sym *sym, const void *key);
+sen_rc sen_sym_del(sen_sym *sym, const void *key);
 unsigned int sen_sym_size(sen_sym *sym);
-int sen_sym_key(sen_sym *sym, sen_id id, void *keybuf, int bufsize);
-sen_set * sen_sym_prefix_search(sen_sym *sym, const char *key);
-sen_set * sen_sym_suffix_search(sen_sym *sym, const char *key);
-sen_id sen_sym_common_prefix_search(sen_sym *sym, const char *key);
+int sen_sym_key(sen_sym *sym, sen_id id, void *keybuf, int buf_size);
+sen_set * sen_sym_prefix_search(sen_sym *sym, const void *key);
+sen_set * sen_sym_suffix_search(sen_sym *sym, const void *key);
+sen_id sen_sym_common_prefix_search(sen_sym *sym, const void *key);
 int sen_sym_pocket_get(sen_sym *sym, sen_id id);
 sen_rc sen_sym_pocket_set(sen_sym *sym, sen_id id, unsigned int value);
 sen_id sen_sym_next(sen_sym *sym, sen_id id);
 
 /******** utility API ********/
-sen_snip *sen_snip_open(sen_encoding encoding, int flags, size_t width, unsigned int max_results,
-                        const char *defaultopentag, const char *defaultclosetag,
+sen_snip *sen_snip_open(sen_encoding encoding, int flags, unsigned int width,
+                        unsigned int max_results,
+                        const char *defaultopentag, unsigned int defaultopentag_len,
+                        const char *defaultclosetag, unsigned int defaultclosetag_len,
                         sen_snip_mapping *mapping);
 sen_rc sen_snip_close(sen_snip *snip);
-sen_rc sen_snip_add_cond(sen_snip *snip, const char *keyword,
-                         const char *opentag, const char *closetag);
-sen_rc sen_snip_exec(sen_snip *snip, const char *string, unsigned int *nresults,
-                     size_t *max_tagged_len);
-sen_rc sen_snip_get_result(sen_snip *snip, const unsigned int index, char *result);
+sen_rc sen_snip_add_cond(sen_snip *snip,
+                         const char *keyword, unsigned int keyword_len,
+                         const char *opentag, unsigned int opentag_len,
+                         const char *closetag, unsigned int closetag_len);
+sen_rc sen_snip_exec(sen_snip *snip,
+                     const char *string, unsigned int string_len,
+                     unsigned int *nresults, unsigned int *max_tagged_len);
+sen_rc sen_snip_get_result(sen_snip *snip, const unsigned int index,
+                           char *result, unsigned int *result_len);
 
 sen_records_heap *sen_records_heap_open(int size, int limit, sen_sort_optarg *optarg);
 sen_rc sen_records_heap_add(sen_records_heap *h, sen_records *r);
@@ -386,17 +401,17 @@ sen_records *sen_records_heap_head(sen_records_heap *h);
 sen_rc sen_records_heap_close(sen_records_heap *h);
 
 int sen_inv_entry_info(sen_inv *inv, sen_id key, unsigned *a, unsigned *pocket,
-		       unsigned *chunk, unsigned *chunk_size, unsigned *buffer_free,
-		       unsigned *nterms, unsigned *nterms_void, unsigned *tid,
-		       unsigned *size_in_chunk, unsigned *pos_in_chunk,
-		       unsigned *size_in_buffer, unsigned *pos_in_buffer);
+                       unsigned *chunk, unsigned *chunk_size, unsigned *buffer_free,
+                       unsigned *nterms, unsigned *nterms_void, unsigned *tid,
+                       unsigned *size_in_chunk, unsigned *pos_in_chunk,
+                       unsigned *size_in_buffer, unsigned *pos_in_buffer);
 
 /* misc */
 
 sen_rc sen_logger_info_set(const sen_logger_info *info);
 
 void sen_logger_put(sen_log_level level,
-		    const char *file, int line, const char *func, char *fmt, ...);
+                    const char *file, int line, const char *func, char *fmt, ...);
 
 int sen_logger_pass(sen_log_level level);
 
@@ -406,7 +421,7 @@ if (sen_logger_pass(level)) {\
 }
 
 #ifndef SEN_LOG_DEFAULT_LEVEL
-#define SEN_LOG_DEFAULT_LEVEL sen_log_info
+#define SEN_LOG_DEFAULT_LEVEL sen_log_notice
 #endif /* SEN_LOG_DEFAULT_LEVEL */
 
 #define sen_log(...) \
