@@ -171,7 +171,7 @@ calc_dsize2(int lrid, int rid, int lsid, int sid, int tf, int score, int psize)
 }
 
 static inline void
-inv_inspect(sen_inv *inv)
+inv_inspect(sen_inv *inv, sen_sym *lexicon)
 {
   int n_terms = 0;
   int n_terms_with_hits = 0;
@@ -184,12 +184,11 @@ inv_inspect(sen_inv *inv)
   int total_nposts = 0;
   sen_id tid;
   sen_inv_cursor *c;
-  sen_sym *lexicon = inv->lexicon;
   if (verbose) { puts("     tid,      df,    nposts,      size   |  term"); }
   for (tid = 1; tid <= lexicon->header->nrecords; tid++) {
     int ps, df = 0, nposts = 0, size = 0, dsize = 0, psize = 0, dsize1 = 0, dsize2 = 0;
     n_terms++;
-    if (!(c = sen_inv_cursor_open(inv, tid))) { continue; }
+    if (!(c = sen_inv_cursor_open(inv, tid, 1))) { continue; }
     n_terms_with_hits++;
     if (c->buf) {
       uint8_t *pp;
@@ -298,7 +297,7 @@ index_inspect(sen_index *i)
   printf("file_size_lexicon  %24u\n", file_size_lexicon);
   printf("inv_segment_size   %24u\n", inv_seg_size);
   printf("inv_chunk_size     %24u\n", inv_chunk_size);
-  inv_inspect(i->inv);
+  inv_inspect(i->inv, i->lexicon);
 }
 
 int
